import base64
import os
import re
import time
from ctypes import (
    Array,
    addressof,
    c_char,
    c_long,
    create_string_buffer
)
from datetime import datetime
from logging import Logger
from threading import Lock
from xml.etree import cElementTree as eTree

from cfgtools import Group
from messagebus import DataType, DefaultToken, Message
from mbcontextmessagehandler import MbContextMessageBus
from modelcommon.customexception import SitefUnavailableException
from modelcommon.model import SitefResponse
from modelcommon.sitef import (
    SitefFunctions,
    ResultCodes,
    CommandCodes,
    FieldCodes,
    FinalizeActionFunction,
    CancelCodes
)
from pos_model import TenderType
from typing import (
    Any,
    List,
    Optional,
    Union
)

from ._SitefActions import SitefActions
from ._SitefParams import SitefParams
from ._SitefServiceFinder import SitefServiceFinder
from dialogmanager import DialogManager
from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    FinishPaymentData,
    PaymentResponse,
    PaymentResponseData,
    PaymentResponseMessage,
    StartPaymentData,
    TestConnectionData
)
from util import get_transaction_identifier

TK_STORECFG_GET = 0x01300001
TK_SLCTRL_OMGR_ORDERPICT = 0x00A00002

API_SUCCESS_OK = 0
API_LOOP_OK = 10000
API_LOOP_FINISHED = 0
API_NO_SITEF_COMMUNICATION = -5


def _get_terminal_id(
    pdv_tef     # type: int
):
    # type: (...) -> str

    return "SW%06d" % pdv_tef


def _api_has_data(command):
    return command.value == API_SUCCESS_OK


def _format_api_error_response(api_ret):
    error_responses = {
        -2: PaymentResponseMessage("Cancelado"),
        -5: PaymentResponseMessage("Falha no Servidor SiTEF"),
        -6: PaymentResponseMessage("Operacao cancelada pelo usuario"),
        -40: PaymentResponseMessage("Nao Autorizado - SiTEF"),
        -43: PaymentResponseMessage("Erro PinPad"),
        -120: PaymentResponseMessage("Cancelado")
    }

    if api_ret > 0:
        return PaymentResponseMessage("Nao Autorizado")

    return error_responses.get(api_ret, PaymentResponseMessage("Falha na Operacao: Erro {}".format(api_ret)))


def _should_ignore_command(command):
    return command.value in (2, 4, 21, 14, 52)


def _get_dll_path(config):
    dll_path = config.find_value("DLLPath", "")
    for var in re.findall("{[a-zA-Z0-9]*}", dll_path):
        dll_path = dll_path.replace(var, os.environ.get(var.replace('{', '').replace('}', ''), ""))

    return os.path.normpath(os.path.join(dll_path, "clisitef32i.dll"))


class SitefPaymentProcessor(object):
    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        dialog_manager,     # type: DialogManager
        config,             # type: Group
        logger              # type: Logger
    ):
        # type: (...) -> None

        from ctypes import WinDLL
        self.logger = logger
        self.dialog_manager = dialog_manager
        self.message_bus = message_bus

        self.store_id = config.find_value("IdLoja")
        self.timeout = int(config.find_value("Timeout", 60))
        self.show_qr_code = config.find_value("ShowQRCode")
        self.is_sync = config.find_value("IsMobileSynch", "False") == "True"
        self.pdv_tef = int(config.find_value("PdvTef") or config.find_value("PosId"))

        dll_path = _get_dll_path(config)
        self.sitef_dll = WinDLL(dll_path)
        self.configura_sitef_interativo = self.sitef_dll["ConfiguraIntSiTefInterativo"]
        self.inicia_sitef_interativo = self.sitef_dll["IniciaFuncaoSiTefInterativo"]
        self.continua_sitef_interativo = self.sitef_dll["ContinuaFuncaoSiTefInterativo"]
        self.finaliza_sitef_interativo = self.sitef_dll["FinalizaFuncaoSiTefInterativo"]

        self.service_finder = SitefServiceFinder(
            self.timeout,
            config.find_values("IP") or [],
            self.store_id,
            self.pdv_tef,
            self.logger,
            dll_path
        )

        self.current_ip = None
        self.current_order = None

        self.cancel = False
        self.cancel_lock = Lock()

    def stop(self):  # type: () -> None
        self.service_finder.stop_search_thread()

    def connection_status(
        self,
        test_connection_data    # type: TestConnectionData
    ):
        # type: (...) -> bool

        return self.service_finder.process(
            pos_id=test_connection_data.pos_id,
            data_fiscal=test_connection_data.fiscal_date,
            hora_fiscal=test_connection_data.fiscal_time
        )

    def add_payment(self, start_payment_data, additional_params=""):
        self.service_finder.lock_sitef()
        try:
            result = self._process_loop(start_payment_data, additional_params)
        finally:
            self.dialog_manager.close_dialogs(start_payment_data.pos_id)
            self.service_finder.release_sitef()

        return result

    def cancel_payment_order(
        self,
        cancel_payment_order_data   # type: CancelPaymentOrderData
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        try:
            self.service_finder.lock_sitef()
            return self.execute_cancel_payment_order(cancel_payment_order_data)
        finally:
            pos_id = cancel_payment_order_data.pos_id
            self.dialog_manager.close_dialogs(pos_id)
            self.service_finder.release_sitef()

    def execute_cancel_payment_order(
        self,
        cancel_payment_order_data    # type: CancelPaymentOrderData
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        pos_id = cancel_payment_order_data.pos_id
        tender_type = cancel_payment_order_data.tender_type
        tender_amount = cancel_payment_order_data.tender_amount
        order_id = cancel_payment_order_data.order_id
        fiscal_date = str(cancel_payment_order_data.fiscal_date)
        fiscal_time = str(cancel_payment_order_data.fiscal_time)
        tender_id = cancel_payment_order_data.tender_id

        self.logger.debug("SitefProcessor Cancel payment order - STARTED")

        self._check_pin_pad(pos_id)
        self._check_sitef_is_available()
        self._notify_start_process(
            pos_id=int(pos_id),
            tender_type=str(tender_type)
        )

        self.cancel = False
        type_code = SitefFunctions.get_cancel_code(tender_type=tender_type)
        amount = "{:,.2f}".format(float(tender_amount)).replace(",", "").replace(".", ",")
        transaction_identifier = get_transaction_identifier(
            order_id=order_id,
            tender_id=tender_id
        )
        result = self.inicia_sitef_interativo(
            type_code,
            amount,
            transaction_identifier,
            fiscal_date,
            fiscal_time,
            "",
            ""
        )
        document_to_cancel = self._get_document_to_cancel(
            type_code,
            cancel_payment_order_data
        )

        if result == API_LOOP_OK:
            response = SitefResponse()
            response.transaction_identifier = transaction_identifier
            command = c_long(0)
            field_type = c_long(0)
            min_size = c_long(0)
            max_size = c_long(0)
            buffer_size = 32768
            output_buffer = create_string_buffer('\000' * buffer_size)

            last_time = -1
            last_message = ""

            show_message_commands = (
                SitefActions.SHOW_OPERATOR_MESSAGE,
                SitefActions.SHOW_OPERATOR_AND_CUSTOMER_MESSAGE,
            )

            self.current_order = order_id
            while result == API_LOOP_OK:
                with self.cancel_lock:
                    if self.cancel:
                        keep_going = -1 if tender_type == 122 else 2
                    else:
                        keep_going = 0

                result = self.continua_sitef_interativo(
                    addressof(command),
                    addressof(field_type),
                    addressof(min_size),
                    addressof(max_size),
                    output_buffer,
                    buffer_size,
                    keep_going
                )

                if result != API_LOOP_OK:
                    break

                if not _api_has_data(command):
                    self.logger.debug("COMMAND TYPE: {0} - BUFFER: {1}".format(command.value, output_buffer.value))

                    if command.value in show_message_commands:
                        key_cancel = "$INSERT_CARD_CANCEL_PAYMENT|{}".format(cancel_payment_order_data.card_number)
                        data = (command, output_buffer, key_cancel, None)
                        last_message = self._call_command_function(*data)
                        if "insira" not in str(output_buffer.value).lower():
                            self.dialog_manager.show_message(pos_id, output_buffer.value)
                    elif command.value == CommandCodes.set_amount_transaction:
                        output_buffer.value = amount
                    elif self._idle_field(command, field_type, FieldCodes.cancel_transaction_date):
                        output_buffer.value = datetime.strptime(fiscal_date, '%Y%m%d').strftime('%d%m%Y')
                    elif self._idle_field(command, field_type, FieldCodes.cancel_document_number):
                        output_buffer.value = str(document_to_cancel)
                    elif command.value == CommandCodes.idle_command:
                        data = (pos_id, order_id, last_time, output_buffer, max_size, last_message)
                        last_time = self._wait_action(*data)
                    elif command.value == SitefActions.MENU_CONTENT and type_code == CancelCodes.digital_wallet:
                        choices = {option.split(':')[1]: option.split(':')[0] for option
                                   in output_buffer.value.split(';') if option}
                        output_buffer.value = str(choices[cancel_payment_order_data.media])

                if field_type.value != FieldCodes.empty_information:
                    self.logger.debug("FIELD TYPE: {0} - BUFFER: {1}".format(field_type.value, output_buffer.value))
                    response = self._set_response_cancel_data(
                        response=response,
                        field_type=field_type.value,
                        output_buffer=output_buffer
                    )

            if result == ResultCodes.success:
                self._notify_end_process(pos_id=pos_id)
                confirm = FinalizeActionFunction.confirm
                self.finaliza_sitef_interativo(
                    confirm,
                    transaction_identifier,
                    fiscal_date,
                    fiscal_time,
                    None
                )
                self.logger.debug("SitefProcessor Cancel payment order - SUCCESS")
                return response

            cancel = FinalizeActionFunction.cancel
            self.finaliza_sitef_interativo(
                cancel,
                transaction_identifier,
                fiscal_date,
                fiscal_time,
                None
            )
            self._notify_end_process(pos_id=pos_id)

            return "$ERROR_CANCEL_ELECTRONIC_PAYMENTS_ORDER"

    @staticmethod
    def _get_document_to_cancel(
        type_code,                  # type: int
        cancel_payment_order_data   # type: CancelPaymentOrderData
    ):
        # type: (...) -> str

        auth_code = cancel_payment_order_data.auth_code
        nsu_host = cancel_payment_order_data.nsu_host

        return auth_code if type_code == CancelCodes.digital_wallet else nsu_host

    def cancel_payment(self):
        with self.cancel_lock:
            self.cancel = True
        return ""

    def finish_payment(
        self,
        finish_payment_data     # type: FinishPaymentData
    ):
        # type: (...) -> bool

        self.logger.debug("SiTEF finish payment started")
        order_id = finish_payment_data.order_id
        pos_id = finish_payment_data.pos_id
        tender_id = finish_payment_data.tender_id
        status = int(finish_payment_data.status)
        transaction_identifier = get_transaction_identifier(
            order_id=order_id,
            tender_id=tender_id
        )
        try:
            self.service_finder.lock_sitef()
            if not self.current_order:
                self.logger.debug("No order to finish, ignoring...")
                return True

            if self.current_order != order_id:
                self.logger.error("order to finish = {}, current order = {}".format(order_id, self.current_order))
                return True

            self.logger.debug("will finish with data: {}".format(finish_payment_data))

            self.configura_sitef_interativo(
                str(self.current_ip),
                str(self.store_id),
                _get_terminal_id(self.pdv_tef),
                0
            )
            fiscal_date = str(finish_payment_data.fiscal_date)
            fiscal_time = str(finish_payment_data.fiscal_time)
            self.finaliza_sitef_interativo(status, transaction_identifier, fiscal_date, fiscal_time, None)

            self.current_ip = self.current_order = None
            self.logger.debug("SiTEF finish payment ended")
            return True
        except Exception as _:
            message = "SITEF Finalize - Pos: {0} / Order: {1} / Status: {2} / Coupon: {3}".format(
                pos_id,
                order_id,
                status,
                transaction_identifier
            )
            self.logger.exception(message)
            return False
        finally:
            self.service_finder.release_sitef()

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        pass

    def _check_finalized_last_order(self, pos_id):
        for _ in xrange(0, 3, 1):
            if self.current_ip and self.current_order:
                order_id = self.current_order
                order = self._get_order_picture(pos_id, order_id)
                if not order:
                    continue
                order = eTree.XML(order).find("Order")
                created_at = order.get("createdAt").replace("-", "").replace(":", "")
                data_fiscal = created_at[:8]
                hora_fiscal = created_at[9:15]
                if order.find("StateHistory/State[@state='PAID']") is not None:
                    status = "1"
                elif order.find("StateHistory/State[@state='VOIDED']") is not None:
                    status = "0"
                else:
                    self.logger.error("Order %s sem status PAID ou VOIDED" % order_id)
                    return
                self.finish_payment(FinishPaymentData(pos_id, order_id, data_fiscal, hora_fiscal, status, ""))
            else:
                return

        self.logger.warning("Impossivel finalizar order %s" % self.current_order)
        self.current_order = self.current_ip = None

    def _get_order_picture(self, pos_id, order_id):
        try:
            pos_id_to_search = pos_id
            if order_id not in ('', None, 0):
                pos_id_to_search = ''
            params = (pos_id_to_search, order_id, 'false')
            reqbuf = '\x00'.join((str(p) for p in params))
            msg = self.message_bus.send_message("ORDERMGR%s" % pos_id, Message(TK_SLCTRL_OMGR_ORDERPICT, DataType.param, reqbuf))
            params = msg.data.split('\x00')
            if msg.token == DefaultToken.TK_SYS_NAK:
                return None
            return params[2]
        except Exception: # noqa
            self.logger.exception("Exception trying to get last order picture")
            return None

    def get_name(self):
        return "Sitef"

    def _check_pin_pad(
        self,
        pos_id,                 # type: int
        electronic_type=-1,     # type: int
    ):
        # type: (...) -> None

        if electronic_type == TenderType.digital_wallet:
            return

        self.dialog_manager.show_message(pos_id, "Conectando PinPad")
        time_limit = (time.time() + 5.0)
        while (time.time() < time_limit) and (not self.sitef_dll["VerificaPresencaPinPad"]()):
            time.sleep(0.2)

    def _process_loop(
        self,
        sitef_loop_data,         # type: StartPaymentData
        additional_params=""     # type: str
    ):
        # type: (...) -> PaymentResponse

        self.logger.debug("SiTEF loop started")
        self.cancel = False
        pos_id = sitef_loop_data.pos_id
        order_id = sitef_loop_data.order_id
        ip_tef = sitef_loop_data.ip_tef
        tender_id = sitef_loop_data.tender_id

        self._check_pin_pad(pos_id, sitef_loop_data.electronic_type)

        if self.current_order == order_id:
            ip_sitef = self.current_ip
        else:
            ip_sitef = self.service_finder.find_sitef_service() if ip_tef is None else ip_tef
            self._check_finalized_last_order(pos_id)

        if not ip_sitef:
            self.service_finder.wakeup_search()
            self.logger.warning("No IP available")
            raise SitefUnavailableException("Sem conexao SITEF")

        self._check_sitef_is_available()

        amount = sitef_loop_data.amount

        electronic_type = sitef_loop_data.electronic_type
        sitef_tender_type = None
        if electronic_type in (TenderType.debit, TenderType.voucher):
            sitef_tender_type = 2
        elif sitef_loop_data.electronic_type == TenderType.credit:
            sitef_tender_type = 3
        elif sitef_loop_data.electronic_type == TenderType.digital_wallet:
            sitef_tender_type = 122
        elif electronic_type == TenderType.connection_test:
            sitef_tender_type = 111

        transaction_identifier = get_transaction_identifier(
            order_id=order_id,
            tender_id=tender_id
        )
        fiscal_date = str(sitef_loop_data.fiscal_date)
        fiscal_time = str(sitef_loop_data.fiscal_time)
        operator = str(sitef_loop_data.operator)
        disabled_payments = "[10;17;18;19;27;28;29;34;35;36;42;]"
        api_ret = self.inicia_sitef_interativo(
            sitef_tender_type,
            amount,
            transaction_identifier,
            fiscal_date, fiscal_time,
            operator, "{};{}".format(disabled_payments, additional_params))

        if api_ret != API_LOOP_OK:
            self._log_api_start_ret_error(api_ret)
            return PaymentResponseMessage("Pagamento interrompido")

        self.current_order = order_id
        self.current_ip = ip_sitef

        command = c_long(0)
        field_type = c_long(0)
        min_size = c_long(0)
        max_size = c_long(0)
        buffer_size = 32768
        output_buffer = create_string_buffer('\000' * buffer_size)

        voided = False

        institution_name = ""
        card_bin = None
        transaction_date = None
        auth_code = None
        authorizer_name = None
        authorizer_cnpj = None
        card_type = None
        expiration_date = None
        nsu = None
        nsu_host = None
        transaction_processor = None
        receipt_merchant = None
        receipt_customer = None
        id_mercado_pago = None
        menu_title = ""
        last_digits = "0000"
        sat_acquirer = ""
        nfce_brand = ""

        show_message_commands = (
            SitefActions.SHOW_OPERATOR_MESSAGE,
            SitefActions.SHOW_OPERATOR_AND_CUSTOMER_MESSAGE,
        )

        while api_ret == API_LOOP_OK:
            keep_going = 0
            with self.cancel_lock:
                if self.cancel:
                    keep_going = -1 if electronic_type == 122 else 2

            api_ret = self.continua_sitef_interativo(
                addressof(command),
                addressof(field_type),
                addressof(min_size),
                addressof(max_size),
                output_buffer,
                buffer_size,
                keep_going
            )
            self.logger.debug("APRET: {0}".format(str(api_ret)))
            if api_ret != API_LOOP_OK:
                break

            self.logger.debug("COMMAND TYPE: {0} - BUFFER: {1}".format(command.value, output_buffer.value))
            if not _api_has_data(command) and not self._must_ignore_command(field_type):
                if command.value in show_message_commands:
                    self.dialog_manager.show_message(pos_id, output_buffer.value)
                elif command.value == SitefActions.SHOW_MESSAGE_WITH_CONFIRMATION:
                    self.dialog_manager.show_question(pos_id, output_buffer.value, buttons="$OK")
                elif command.value == SitefActions.SHOW_MESSAGE_WITH_CANCEL_BUTTON:
                    self.dialog_manager.show_message(pos_id, output_buffer.value)
                elif command.value == SitefActions.ASK_PIN_CODE:
                    pin = self.dialog_manager.show_input(pos_id, output_buffer.value, input_type="number")
                    if pin == "CANCELA":
                        self.cancel_payment()
                        output_buffer.value = "0" * max_size.value
                    else:
                        output_buffer.value = str(pin)
                elif command.value == SitefActions.CANCEL_OPERATION:
                    output_buffer.value = "0"
                    time.sleep(1.5)

                    if field_type.value != 111:
                        voided = True
                elif command.value == SitefActions.MENU_TITLE:
                    menu_title = output_buffer.value
                elif command.value == SitefActions.CLEAN_MENU_TITLE:
                    menu_title = ""
                elif command.value == SitefActions.MENU_CONTENT:
                    choices = {option.split(':')[0]: option.split(':')[1] for option in output_buffer.value.split(';') if option}
                    selected_choice = self.dialog_manager.show_choices(pos_id, choices, message=menu_title)
                    output_buffer.value = str(selected_choice)
                elif command.value == SitefActions.CLOSE_QRCODE:
                    self.dialog_manager.close_qrcode(pos_id)
                elif _should_ignore_command(command):
                    time.sleep(0.5)
                else:
                    self.dialog_manager.show_message(pos_id, "Mensagem {} {}".format(command.value, output_buffer.value))

            if field_type.value == SitefParams.CARD_BIN:
                card_bin = output_buffer.value
            elif field_type.value == SitefParams.TRANSACTION_PROCESSOR:
                transaction_processor = output_buffer.value
            elif field_type.value == SitefParams.CARD_TYPE:
                card_type = output_buffer.value
            elif field_type.value in SitefParams.AUTH_CODE:
                auth_code = output_buffer.value
            elif self._is_show_qrcode_command(field_type):
                self.dialog_manager.show_qrcode(pos_id, output_buffer.value)
            elif field_type.value == SitefParams.RECEIPT_CUSTOMER:
                receipt_customer = output_buffer.value
            elif field_type.value == SitefParams.RECEIPT_MERCHANT:
                receipt_merchant = output_buffer.value
            elif field_type.value == SitefParams.NSU:
                nsu = output_buffer.value
            elif field_type.value == SitefParams.NSU_HOST:
                nsu_host = output_buffer.value
            elif field_type.value == SitefParams.TRANSACTION_DATE:
                transaction_date = output_buffer.value
            elif field_type.value == SitefParams.AUTHORIZER_NAME:
                authorizer_name = output_buffer.value
            elif field_type.value == SitefParams.INSTITUTION_NAME:
                institution_name = output_buffer.value
            elif field_type.value == SitefParams.EXPIRATION_DATE:
                expiration_date = output_buffer.value
            elif field_type.value == SitefParams.AUTHORIZER_CNPJ:
                authorizer_cnpj = output_buffer.value
            elif field_type.value == SitefParams.ID_MERCADO_PAGO:
                id_mercado_pago = output_buffer.value
            elif field_type.value == SitefParams.DIGITAL_WALLET_ID:
                card_type = output_buffer.value
            elif field_type.value == SitefParams.LAST_CARD_DIGITS:
                last_digits = output_buffer.value
            elif field_type.value == SitefParams.SAT_ACQUIRER:
                sat_acquirer = output_buffer.value
            elif field_type.value == SitefParams.NFCE_BRAND:
                nfce_brand = output_buffer.value

        if api_ret == API_LOOP_FINISHED:
            self.logger.debug("SiTEF loop finished")
            return PaymentResponseData(
                media=institution_name.strip(),
                result=api_ret,
                card_number=card_bin,
                auth_code="{}{}".format(transaction_date, id_mercado_pago) if electronic_type == 122 else auth_code,
                approved_amount=amount,
                id_auth=authorizer_name,
                cnpj_auth=authorizer_cnpj,
                bandeira=card_type,
                receipt_merchant=base64.b64encode(receipt_merchant or receipt_customer),
                receipt_customer=base64.b64encode(receipt_customer),
                expiration_date=expiration_date,
                nsu=nsu,
                transaction_processor=transaction_processor,
                last_digits=last_digits,
                sat_acquirer=sat_acquirer,
                nfce_brand=nfce_brand,
                original_nfce_brand=nfce_brand,
                nsu_host=nsu_host,
                electronic_type=tender_id,
                transaction_identifier=transaction_identifier
            )
        else:
            if voided:
                return PaymentResponseMessage("Cancelado")

            if api_ret == API_NO_SITEF_COMMUNICATION:
                self.service_finder.sitef_unavailable(ip_sitef)

            return _format_api_error_response(api_ret)

    def _must_ignore_command(self, field_type):
        return self._is_show_qrcode_command(field_type)

    @staticmethod
    def _is_show_qrcode_command(field_type):
        return field_type.value == SitefParams.QR_CODE

    def _log_api_config_ret_error(self, api_ret):
        error_codes = {
            0: "No Error",
            1: "IP Error",
            2: "Store Code Error",
            3: "Terminal Code Error",
            6: "TCP/IP Error",
            7: "Memory Error",
            8: "CliSiTef Error",
            10: "CliSiTef Access Error"
        }
        self.logger.error("API Config Error: {}".format(error_codes.get(api_ret, "Unknown")))

    def _log_api_start_ret_error(self, api_ret):
        self.logger.warning("API Start Error: {}".format(api_ret))

    @staticmethod
    def _get_coupon_code(
        pos_id,     # type: int
        order_id,   # type: int
        tender_id   # type: str
    ):
        return "{0}{1}{2}".format(pos_id, order_id, tender_id)

    def _check_sitef_is_available(
        self
    ):
        # type: (...) -> None

        self.current_ip = self.service_finder.find_sitef_service() if self.current_ip is None else self.current_ip
        result = self.configura_sitef_interativo(
            str(self.current_ip),
            str(self.store_id),
            _get_terminal_id(self.pdv_tef)
            ,
            0
        )
        if result != API_SUCCESS_OK:
            self.service_finder.sitef_unavailable(self.current_ip)
            self._log_api_config_ret_error(result)
            raise SitefUnavailableException("$SITEF_UNAVAILABLE")

    def _notify_start_process(
        self,
        pos_id,         # type: int
        tender_type     # type: str
    ):
        # type: (...) -> None

        self.message_bus.mbcontext.MB_EasyEvtSend(
            subject="SITEF",
            type="PROCESS_STARTED",
            xml=tender_type,
            sourceid=pos_id,
            synchronous=True
        )

    def _notify_status_update(
        self,
        pos_id,     # type: Any[Union[str, int]]
        data=""     # type: Optional[str]
    ):
        # type: (...) -> None

        data = ";".join([data, pos_id])
        self.message_bus.mbcontext.MB_EasyEvtSend(
            subject="SITEF_STATUS_UPDATE",
            type="STATUS_UPDATE",
            xml=data
        )

    def _notify_end_process(
        self,
        pos_id  # type: int
    ):
        # type: (...) -> None

        self.message_bus.mbcontext.MB_EasyEvtSend(
            subject="SITEF_STATUS_UPDATE_FINISHED",
            type="SITEF_STATUS_UPDATE_FINISHED",
            xml=str(pos_id)
        )

    @staticmethod
    def _idle_field(
        command,        # type: c_long
        field_type,     # type: c_long,
        value           # type: int
    ):
        # type: (...) -> bool

        return command.value == CommandCodes.idle_field and field_type.value == value

    def _wait_action(
        self,
        pos_id,         # type: str
        order_id,       # type: str
        last_time,      # type: int
        output_buffer,  # type: Array[c_char]
        max_size,       # type: c_long
        last_message    # type: str
    ):
        # type: (...) -> float

        if last_time == -1:
            return time.time()

        current_time = time.time()
        if current_time - last_time > self.timeout:
            with self.cancel_lock:
                self.cancel = True

            log_message = "SyncSitefProcessor - Wait Action - OrderId: {0} - IP SiTef: {1}"
            self.logger.debug(log_message.format(order_id, self.current_ip))

            output_buffer.value = "0" * max_size.value
            return last_time

        delimiter = "|" if last_message.startswith("$") else " "
        message = "{}{}{}".format(last_message, delimiter, str(self.timeout - int(current_time - last_time)))
        self._notify_status_update(
            pos_id=pos_id,
            data=message
        )

        return last_time

    def _call_command_function(
        self,
        command,                    # type: c_long
        output_buffer,              # type: Array[c_char]
        key_translate=None,         # type: str
        extra_data_display=None     # type: Optional[List[str]]
    ):
        # type: (...) -> str

        if command.value in [CommandCodes.show_message_operator, CommandCodes.show_message_both]:
            return self.format_message_translate(
                output_buffer=output_buffer,
                key_translate=key_translate,
                extra_data_display=extra_data_display
            )

        if command.value in [CommandCodes.clear_message_operator, CommandCodes.clear_message_both]:
            # TODO: In SITEF here was a call to SitefCallback
            return ""

        return ""

    @staticmethod
    def format_message_translate(
        output_buffer,              # type: Array[c_char]
        key_translate,              # type: Optional[str]
        extra_data_display=None     # type: Optional[List[str]]
    ):
        # type: (...) -> str

        if key_translate is not None and extra_data_display is None:
            return key_translate

        if extra_data_display is None:
            return output_buffer.value

        extra_data_display.insert(0, output_buffer.value)
        message = "|".join([str(e) for e in extra_data_display])

        return "{}|{}".format(key_translate, message)

    @staticmethod
    def _set_response_cancel_data(
        response,       # type: SitefResponse
        field_type,     # type: int
        output_buffer   # type: Array[c_char]
    ):
        # type: (...) -> SitefResponse

        if field_type == FieldCodes.receipt_customer:
            response.customer_receipt = base64.b64encode(output_buffer.value)
        elif field_type == FieldCodes.receipt_merchant:
            response.merchant_receipt = base64.b64encode(output_buffer.value)
        elif field_type == FieldCodes.nsu:
            response.nsu = output_buffer.value
        elif field_type == FieldCodes.nsu_host:
            response.nsu_host = output_buffer.value
        elif field_type == FieldCodes.nsu_origin:
            response.nsu_origin = output_buffer.value
        elif field_type == FieldCodes.transaction_value:
            response.transaction_value = output_buffer.value

        return response



from typing import Dict, Tuple, Optional, List


class FiscalParameterNotFound(Exception):
    def __init__(self, part_code, parameter_name):
        # type: (str, unicode) -> None

        super(FiscalParameterNotFound, self).__init__("FiscalParameterNotFound")
        self.part_code = str(part_code)
        self.parameter_name = parameter_name

    def __str__(self):
        return "FiscalParameterNotFound: partCode: {0} - parameterName: {1}".format(self.part_code, self.parameter_name)


class FiscalParameterController(object):
    def __init__(self, fiscal_parameters, tip_config_range, loader_fiscal_parameters):
        self.fiscal_parameters = fiscal_parameters  # type: Dict[Tuple[str, unicode],unicode]
        self.tip_config_range = tip_config_range  # type: Dict
        self.loader_fiscal_parameters = loader_fiscal_parameters  # type: Dict

    def get_optional_parameter(
            self,
            part_code,      # type: str
            parameter_name  # type: unicode
    ):
        # type: (...) -> Optional[unicode]

        key = (str(part_code), parameter_name)
        if key not in self.fiscal_parameters:
            return None
        else:
            return self.fiscal_parameters[key] or None

    def get_parameter(
            self,
            part_code,      # type: str
            parameter_name  # type: unicode
    ):
        # type: (...) -> unicode

        key = (str(part_code), parameter_name)
        if key not in self.fiscal_parameters:
            raise FiscalParameterNotFound(part_code, parameter_name)
        else:
            return self.fiscal_parameters[key] or None

    def has_parameter(
            self,
            part_code,      # type: str
            parameter_name  # type: unicode
    ):
        # type: (...) -> bool

        key = (str(part_code), parameter_name)
        return key in self.fiscal_parameters

    def get_all_products_csts(self):
        # type: () -> unicode

        return str({(x, y): self.fiscal_parameters.get((x, y)) for x, y in self.fiscal_parameters if "CST" in y})

    def get_all_products_base_reductions(self):
        # type: () -> unicode

        return str({x: self.fiscal_parameters.get((x, "BASE_REDUCTION")) for x, y in self.fiscal_parameters})

    def get_tip_params(self, part_code=None, taxes_name=None):
        # type: (str, List) -> []

        products = []
        try:
            if len(self.tip_config_range) == 0:
                return products
            for item in self.tip_config_range:
                if item["exempt"] or (part_code is not None and part_code != item["TipCode"]):
                    continue
                products.extend(self._get_taxes(item, taxes_name))
        except Exception as _:
            pass
        return products

    def _get_taxes(self, item, taxes_name=None):
        # type:(Dict, []) -> []

        taxes = ['PIS', 'COFINS', 'ICMS'] if taxes_name is None else taxes_name
        products = []
        for product in self.loader_fiscal_parameters:
            product_code = str(product['PRODUCT_CODE'])
            tip_code = str(item["TipCode"])
            if product_code != tip_code:
                continue
            for param in product:
                if param not in taxes:
                    continue
                rate = float(product[param])
                code = "{}_{}".format(str(param), rate)
                name = str(param)
                product_code = {tip_code: tip_code}
                index = str(param[:1])
                result = {"fiscalIndex": index, "code": code, "rate": rate, "name": name, "product": product_code}

                products.append(result)

        return products



# -*- coding: utf-8 -*-

import base64
import bisect
import glob
import json
import logging
import os
import os.path
import re
import shutil
import zipfile
import zlib
from datetime import datetime, timedelta
from decimal import Decimal as D
from xml.etree import cElementTree as eTree

import fiscalpersistence
import printer
from bustoken import TK_FISCALWRAPPER_PROCESS_REQUEST, TK_FISCALWRAPPER_SEARCH_REQUEST, TK_FISCALWRAPPER_REPRINT, \
    TK_FISCALWRAPPER_SEFAZ_CONNECTIVITY, TK_FISCALWRAPPER_SAT_OPERATIONAL_STATUS_REQUEST, TK_SAT_STATUS_REQUEST, \
    TK_FISCALWRAPPER_EXPORT_FISCAL_FILES, TK_FISCALWRAPPER_GET_NF_TYPE, TK_SAT_OPERATIONAL_STATUS_REQUEST, \
    TK_FISCALWRAPPER_SAT_PROCESS_PAYMENT, TK_FISCALWRAPPER_SAT_PAYMENT_STATUS, TK_SAT_PROCESS_PAYMENT, \
    TK_SAT_PAYMENT_STATUS, TK_FISCALWRAPPER_CANCEL_ORDER, TK_FISCALWRAPPER_RE_SIGN_XML, \
    TK_FISCALWRAPPER_SITUATION, TK_FISCALWRAPPER_GET_CST, TK_FISCALWRAPPER_GET_BASE_REDUCTION, \
    TK_FISCALWRAPPER_GET_FISCAL_XML, TK_FISCALWRAPPER_GET_IBPT_TAX, TK_FISCALWRAPPER_GET_ORDERS_DATA, \
    TK_FISCALWRAPPER_GET_LAST_ORDER_ID, TK_FISCALWRAPPER_GET_CERTIFICATE_EXPIRATION_DAYS, \
    TK_FISCALWRAPPER_GET_CONTINGENCY_STATUS, TK_FISCALWRAPPER_CHANGE_CONTINGENCY_STATUS, \
    TK_FISCALWRAPPER_PRINT_DONATION_COUPON, TK_FISCALWRAPPER_GET_TIP
from common import FiscalParameterController
from comp_exceptions import EmptySearchException, VoidOrderException
from fiscalinterface import FiscalProcessor
from fiscalpaymentparser import convert_order_tenders_to_fiscal_tenders
from fiscalpersistence import FiscalDataRepository
from old_helper import OrderTaker, round_half_away_from_zero, remove_xml_namespace, convert_from_utf_to_localtime
from ibpttaxprocessor import IbptTaxProcessor
from messagehandler import EventHandler
from msgbus import TK_SYS_NAK, TK_SYS_ACK, MBEasyContext, MBException, FM_PARAM, MBMessage, FM_STRING
from nfce import NfceRequestBuilder, NfceConnectivityTester, NfceSituationChecker, NfceContingencia, NfceProcessor
from nfcecanceler import NfceCanceler
from nfcedisabler import OrderDisabler
from order_retriever import OrderRetrieverService
from persistence import Driver as DBDriver
from poslistener import get_updated_sale_line_defaults
from printer import DonationCouponPrinter
from sat import SatProcessor, SatServiceFinder
from services import SequencerService
from sysactions import get_model, get_posot, get_podtype
from systools import sys_log_exception
from typing import Optional, Union, Dict


logger = logging.getLogger("FiscalWrapper")
logger_coupons = logging.getLogger("FiscalWrapperCoupons")


class FiscalWrapperEventHandler(EventHandler):
    def __init__(self, mbcontext, fiscal_processor, printers, nfce_connectivity_tester, nfce_situation_checker, order_disabler, nfce_canceler, fiscal_parameter_controller, xml_enviados, xml_erros, export_xml_path, fiscal_sent_dir, nf_type, versao_ws, ibpt_tax_processor, order_retriever_service, xml_inconsistentes, donation_coupon_printer, nfce_contingencia, period_to_retry_orders_with_exception, synchronous_mode, sequencer_service):
        # type: (MBEasyContext, Union[FiscalProcessor, NfceProcessor, SatProcessor], Dict[str, printer.FiscalPrinter], NfceConnectivityTester, NfceSituationChecker, OrderDisabler, NfceCanceler, FiscalParameterController, str, str, str, str, str, int, IbptTaxProcessor, OrderRetrieverService, str, DonationCouponPrinter, NfceContingencia, str, bool, SequencerService) -> None
        super(FiscalWrapperEventHandler, self).__init__(mbcontext)
        self.fiscal_processor = fiscal_processor
        self.printers = printers
        self.nf_type = nf_type  # type: str
        self.nfce_connectivity_tester = nfce_connectivity_tester
        self.nfce_situation_checker = nfce_situation_checker
        self.order_disabler = order_disabler
        self.order_retriever_service = order_retriever_service
        self.xml_enviados = xml_enviados
        self.xml_erros = xml_erros
        self.export_xml_path = export_xml_path
        self.fiscal_sent_dir = fiscal_sent_dir
        self.nfce_canceler = nfce_canceler
        self.versao_ws = versao_ws
        self.fiscal_parameter_controller = fiscal_parameter_controller
        self.ibpt_tax_processor = ibpt_tax_processor
        self.xml_inconsistentes = xml_inconsistentes
        self.donation_coupon_printer = donation_coupon_printer
        self.nfce_contingencia = nfce_contingencia
        self.period_to_retry_orders_with_exception = period_to_retry_orders_with_exception
        self.synchronous_mode = synchronous_mode
        self.sequencer_service = sequencer_service

    def get_handled_tokens(self):
        return [TK_FISCALWRAPPER_PROCESS_REQUEST, TK_FISCALWRAPPER_SEARCH_REQUEST, TK_FISCALWRAPPER_REPRINT,
                TK_FISCALWRAPPER_SEFAZ_CONNECTIVITY, TK_FISCALWRAPPER_SAT_OPERATIONAL_STATUS_REQUEST, TK_FISCALWRAPPER_EXPORT_FISCAL_FILES, TK_FISCALWRAPPER_GET_NF_TYPE,
                TK_SAT_OPERATIONAL_STATUS_REQUEST, TK_FISCALWRAPPER_SAT_PROCESS_PAYMENT,
                TK_FISCALWRAPPER_SAT_PAYMENT_STATUS, TK_SAT_PROCESS_PAYMENT, TK_SAT_PAYMENT_STATUS, TK_FISCALWRAPPER_CANCEL_ORDER, TK_FISCALWRAPPER_RE_SIGN_XML,
                TK_FISCALWRAPPER_SITUATION, TK_FISCALWRAPPER_GET_CST, TK_FISCALWRAPPER_GET_BASE_REDUCTION, TK_FISCALWRAPPER_GET_FISCAL_XML, TK_FISCALWRAPPER_GET_IBPT_TAX,
                TK_FISCALWRAPPER_GET_LAST_ORDER_ID, TK_FISCALWRAPPER_GET_ORDERS_DATA, TK_FISCALWRAPPER_PRINT_DONATION_COUPON, TK_FISCALWRAPPER_GET_CERTIFICATE_EXPIRATION_DAYS,
                TK_FISCALWRAPPER_GET_CONTINGENCY_STATUS, TK_FISCALWRAPPER_CHANGE_CONTINGENCY_STATUS,
                TK_FISCALWRAPPER_GET_TIP]

    def handle_message(self, msg):
        # type: (MBMessage) -> Union[None, bool]
        try:
            if msg.token == TK_FISCALWRAPPER_PROCESS_REQUEST:
                response = self._handle_process_request(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_CERTIFICATE_EXPIRATION_DAYS:
                response = self._handle_get_certificate_expiration()
            elif msg.token == TK_FISCALWRAPPER_REPRINT:
                response = self._handle_process_request(msg, reprint=True)
            elif msg.token == TK_FISCALWRAPPER_SEARCH_REQUEST:
                response = self._handle_search_request()
            elif msg.token == TK_FISCALWRAPPER_SEFAZ_CONNECTIVITY:
                response = self._handle_sefaz_connectivity_request()
            elif msg.token == TK_FISCALWRAPPER_SAT_OPERATIONAL_STATUS_REQUEST:
                response = self._handle_sat_op_status_request(msg)
            elif msg.token == TK_FISCALWRAPPER_EXPORT_FISCAL_FILES:
                response = self._export_fiscal_files(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_NF_TYPE:
                response = self._handle_get_nftype(msg)
            elif msg.token == TK_FISCALWRAPPER_SAT_PROCESS_PAYMENT:
                response = self._handle_sat_process_payment_request(msg)
            elif msg.token == TK_FISCALWRAPPER_SAT_PAYMENT_STATUS:
                response = self._handle_sat_payment_status_request(msg)
            elif msg.token == TK_FISCALWRAPPER_CANCEL_ORDER:
                response = self._handle_cancel_order(msg)
            elif msg.token == TK_FISCALWRAPPER_RE_SIGN_XML:
                response = self._handle_re_sign_xml()
            elif msg.token == TK_FISCALWRAPPER_SITUATION:
                response = self._handle_sefaz_situation(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_FISCAL_XML:
                response = self._handle_fiscal_xml_request(msg.data, True)
            elif msg.token == TK_FISCALWRAPPER_GET_CST:
                response = self._handle_get_csts(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_BASE_REDUCTION:
                response = self._handle_get_base_reductions(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_TIP:
                response = self._handle_get_tip_params(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_IBPT_TAX:
                response = self._handle_get_ibpt_tax(msg)
            elif msg.token == TK_FISCALWRAPPER_PRINT_DONATION_COUPON:
                response = self._handle_print_donation_coupon(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_CONTINGENCY_STATUS:
                response = self._handle_get_contingency_status(msg)
            elif msg.token == TK_FISCALWRAPPER_CHANGE_CONTINGENCY_STATUS:
                response = self._handle_change_contingency_status(msg)
            elif msg.token == TK_FISCALWRAPPER_GET_ORDERS_DATA:
                try:
                    orders_data_list_str = self.order_retriever_service.get_orders_data(msg.data)

                    msg.token = TK_SYS_ACK
                    self.mbcontext.MB_ReplyMessage(msg, data=orders_data_list_str)
                    return
                except Exception as ex:
                    logger.exception("Error handling TK_FISCALWRAPPER_GET_ORDERS_DATA")
                    msg.token = TK_SYS_NAK
                    self.mbcontext.MB_ReplyMessage(msg, data=ex.message)
                    return
            elif msg.token == TK_FISCALWRAPPER_GET_LAST_ORDER_ID:
                try:
                    last_order_id = self.order_retriever_service.get_last_order_id()

                    msg.token = TK_SYS_ACK
                    self.mbcontext.MB_ReplyMessage(msg, data=last_order_id)
                    return
                except Exception as ex:
                    logger.exception("Error handling TK_FISCALWRAPPER_GET_LAST_ORDER_ID")
                    msg.token = TK_SYS_NAK
                    self.mbcontext.MB_ReplyMessage(msg, data=ex.message)
                    return
            else:
                response = (False, "")

            msg.token = TK_SYS_ACK if response[0] else TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg, data=response[1])

        except Exception as ex:
            sys_log_exception("Unexpected exception when handling event {0}. Error message: {1}".format(msg.token, ex.message))
            msg.token = TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg, data=str(ex))
        except:
            sys_log_exception("Unhandled BaseException Message")
            msg.token = TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg, data=str("Unhandled BaseException"))

        return False

    def terminate_event(self):
        self.fiscal_processor.terminate()

    def _handle_get_nftype(self, _):
        # type: (MBMessage) -> (str, str)
        return True, self.nf_type

    def _handle_get_certificate_expiration(self):
        if self.nf_type == "NFCE":
            ret = self.fiscal_processor.do_validation(get_days_to_expiration=True)
            if not ret[0]:
                return False, ret[1]
            return True, ret[1]

    def _handle_process_request(self, msg, reprint=False):
        # type: (MBMessage, Optional[bool]) -> (str, str)

        ret = self.fiscal_processor.do_validation()
        if not ret[0]:
            return False, '\0'.join(map(str, (False, str(ret[1]))))

        if reprint:
            data = msg.data.split('\0')
            posid = data[0]
            orderid = data[1]
            danfe_type = data[2] if len(data) > 2 else ""
            model = get_model(posid)
            pos_ot = get_posot(model)
            order = eTree.XML(pos_ot.orderPicture(posid=posid, orderid=orderid))
            order = order.find("Order")

            # Add default quantity info to order pict
            dict_list = []
            for line in order.findall("SaleLine"):
                line_number = line.get('lineNumber')
                item_id = line.get('itemId')
                part_code = line.get('partCode')
                level = line.get('level')
                get_updated_sale_line_defaults(dict_list, orderid, int(line_number), item_id, part_code, line.get('qty'), int(level), get_podtype(model))
                for corrected_line in dict_list:
                    corrected_line_dict = json.loads(corrected_line)
                    if corrected_line_dict['line_number'] == line_number and \
                       corrected_line_dict['item_id'] == item_id and \
                       corrected_line_dict['part_code'] == part_code and \
                       corrected_line_dict['level'] == level:
                        line_orig_qty = line.get('qty')
                        line.set('addedQty', str(max(0.0, float(line_orig_qty) - float(corrected_line_dict['default_qty']))))
                        line.set('defaultQty', str(float(corrected_line_dict['default_qty'])))

        else:
            data = msg.data.split('\0')
            posid = data[0]
            orderid = data[1] if len(data) > 1 else ""
            danfe_type = data[2] if len(data) > 2 else ""
            model = get_model(posid)
            pos_ot = get_posot(model)
            order = eTree.XML(pos_ot.orderPicture(posid=posid, orderid=orderid))
            order = order.find("Order") if order.find("Order") else order
            if not orderid:
                orderid = order.get("orderId")

            fiscalization_date = order.find("CustomOrderProperties/OrderProperty[@key='FISCALIZATION_DATE']")
            if fiscalization_date is not None:
                return True, ""

        with FiscalDataRepository(self.mbcontext) as fiscal_repository:  # type: fiscalpersistence.FiscalDataRepository
            fiscal_ok = False
            try:
                fiscal_id_property = "FISCAL_ID"
                fiscal_id_node = order.find("CustomOrderProperties/OrderProperty[@key='{}']".format(fiscal_id_property))

                if fiscal_id_node is not None:
                    fiscal_id_node = fiscal_id_node.get("value")

                if not fiscal_id_node:
                    fiscal_id = self.sequencer_service.build_next_fiscal_id()
                    fiscal_id_node = fiscal_id
                    pos_ot.setOrderCustomProperty(fiscal_id_property, fiscal_id)

                tenders = convert_order_tenders_to_fiscal_tenders(order)

                customer_name = order.find("CustomOrderProperties/OrderProperty[@key='CUSTOMER_NAME']")
                if customer_name is not None:
                    customer_name = customer_name.get("value")

                total_taxes = self.ibpt_tax_processor.ibpt_tax_calculator(order)

                if reprint:
                    if self.nf_type not in ("SAT", "MFE"):
                        nfce_fiscal_processor = self.fiscal_processor  # type: NfceProcessor
                        params = {'tenders': tenders, 'order': order, 'customer_name': customer_name, 'url_sefaz': nfce_fiscal_processor.url_sefaz, 'posid': posid, 'total_taxes': total_taxes}
                    else:
                        params = {'tenders': tenders, 'order': order, 'customer_name': customer_name, 'posid': posid, 'total_taxes': total_taxes}
                    report = self._reprint_sat_or_nfce(fiscal_repository.get_xml_request(orderid), fiscal_repository.get_xml_response(orderid), params, danfe_type)
                    if report[0]:
                        return True, report[1]
                    else:
                        return False, report[1]

                message = "Nova order para processamento - Fiscal ID: {} - Order ID: {}".format(
                        str(fiscal_id_node),
                        str(orderid)
                )

                logger.info(message)
                fiscal_data = self.fiscal_processor.request_fiscal(
                        posid,
                        order,
                        tenders,
                        self.nf_type == "PAF",
                        str(fiscal_id_node)
                )
                fiscal_ok = True

                if self.nf_type in ("SAT", "MFE"):
                    sat_fiscal_processor = self.fiscal_processor  # type: SatProcessor
                    sat_used = sat_fiscal_processor.last_sat_used[posid]
                    sat_used = sat_used[3:]
                    response_xml = eTree.tostring(fiscal_data.find("FiscalData"))
                    base64_response = base64.b64encode(response_xml)

                    fiscal_repository.start_save_fiscal_data_thread(posid, order, fiscal_data.find("XmlRequest").text.strip(), sat_used, 1, base64_response)
                    if danfe_type in self.printers:
                        selected_printer = self.printers[danfe_type]
                    else:
                        selected_printer = self.printers[""]

                    report = selected_printer.print_fiscal(posid, order, tenders, fiscal_data.find("XmlRequest"), fiscal_data.find("FiscalData"), total_taxes, customer_name)
                else:
                    request_xml = eTree.XML(fiscal_data[0])
                    response_xml = None
                    xml_response_base64 = None
                    if fiscal_data[1] is not None:
                        response_xml = eTree.XML(fiscal_data[1])
                        xml_response_base64 = base64.b64encode(fiscal_data[1])
                    idx_start = fiscal_data[0].find("<NFe")
                    idx_final = int(str(fiscal_data[0].find("</NFe"))) + 6
                    xml_string = fiscal_data[0][idx_start:idx_final]

                    contingencia = request_xml.find(
                        "{{{0}}}Body/{{{1}}}nfeDadosMsg/{{{2}}}enviNFe/{{{2}}}NFe/{{{2}}}infNFe/{{{2}}}ide/{{{2}}}tpEmis".format(
                            NfceRequestBuilder.NAMESPACE_SOAP,
                            NfceRequestBuilder.NAMESPACE_AUTORIZACAO if self.versao_ws in (1, 3) else NfceRequestBuilder.NAMESPACE_AUTORIZACAO_4,
                            NfceRequestBuilder.NAMESPACE_NFE)).text
                    if not contingencia == "9":
                        xml_string = fiscal_data[2]
                    xml_base64 = base64.b64encode(xml_string)
                    originator_id = order.get("originatorId")[3:]

                    fiscal_repository.start_save_fiscal_data_thread(originator_id, order, xml_base64, "00", 0 if contingencia == "9" else 1, xml_response_base64)

                    if danfe_type in self.printers:
                        report = self.printers[danfe_type].print_fiscal(posid, order, tenders, request_xml, response_xml, total_taxes, customer_name)
                    else:
                        report = ""

                self._print_coupon(report)

                return True, report

            except VoidOrderException as ex:
                fiscal_ok = False
                return False, '\0'.join(map(str, (fiscal_ok, str(ex))))

            except MBException:
                logger.exception("Erro de Impressão")
                return False, '\0'.join(map(str, (fiscal_ok, "Verifique a Impressora de Cupom Fiscal")))

            except Exception as ex:
                logger.exception("Nota Rejeitada pela SEFAZ")
                return False, '\0'.join(map(str, (fiscal_ok, str(ex))))

    def _print_coupon(self, report):
        try:
            logger_coupons.info("\n\n***************************************************************\n\n")
            logger_coupons.info(report)
        except Exception as _:
            logger_coupons.info("Error logging coupon")
        finally:
            logger_coupons.info("\n\n***************************************************************\n\n\n\n")

    def _handle_sefaz_connectivity_request(self):
        if self.nf_type == "NFCE":
            self.nfce_connectivity_tester.logger = logger
            if self.nfce_connectivity_tester.test_connectivity():
                return True, "Conexao com Sefaz OK!"
            else:
                return False, "Nao foi possivel se conectar com a Sefaz."
        else:
            return False, "Loja nao opera com NFCE"

    def _handle_sefaz_situation(self, msg):
        if self.nf_type == "NFCE":
            initial_date, final_date = msg.data.split('\0')
            initial_date = initial_date[:4] + '-' + initial_date[4:6] + '-' + initial_date[6:]
            final_date = final_date[:4] + '-' + final_date[4:6] + '-' + final_date[6:]

            if self.versao_ws == 1:
                namespace_consulta_situacao = NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO
                tag_consulta_nf_result = "nfeConsultaNF2Result"
            elif self.versao_ws == 3:
                namespace_consulta_situacao = NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO_3
                tag_consulta_nf_result = "nfeConsultaNFResult"
            else:
                namespace_consulta_situacao = NfceRequestBuilder.NAMESPACE_CONSULTA_SITUACAO_4
                tag_consulta_nf_result = "nfeResultMsg"

            with FiscalDataRepository(self.mbcontext) as fiscalrepository:
                logger.info("Tratando Evento Consultar Situacao XML")
                conn = None
                try:
                    conn = DBDriver().open(self.mbcontext, service_name="FiscalPersistence")
                    offset = 0
                    while True:
                        sql = """SELECT PosId, OrderId, XMLRequest, OrderPicture
                                  FROM fiscal.FiscalData
                                  WHERE date(DataNota, 'unixepoch', 'localtime') >= '%s'
                                  AND date(DataNota, 'unixepoch', 'localtime') <= '%s'
                                  ORDER BY OrderId ASC LIMIT 500 OFFSET %s""" % (initial_date, final_date, offset)
                        cursor = conn.select(sql)
                        if cursor.rows() == 0:
                            break
                        offset += cursor.rows()
                        eTree.register_namespace('', NfceRequestBuilder.NAMESPACE_NFE)
                        for row in cursor:
                            pos_id, order_id, request, order_picture = map(row.get_entry, ("PosId", "OrderId", "XMLRequest", "OrderPicture"))
                            logger.info("Order %s " % order_id)

                            request_str = base64.b64decode(request + "=" * ((4 - len(request) % 4) % 4))
                            req = eTree.XML(request_str)

                            c_stat = req.find("{{{0}}}protNFe/{{{0}}}infProt/{{{0}}}cStat".format(NfceRequestBuilder.NAMESPACE_NFE))
                            if c_stat is not None:
                                c_stat = c_stat.text

                            if c_stat in ("100", "150"):
                                logger.info("Order %s ja estava com status 100" % order_id)
                                continue
                            try:
                                logger.info("Buscando Protocolo. Order %s " % order_id)

                                nfe = req.find("{{{0}}}NFe/{{{0}}}infNFe".format(NfceRequestBuilder.NAMESPACE_NFE))
                                nfe_key = nfe.attrib["Id"]
                                nfe = nfe_key[3:]
                                data_emissao = req.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}dhEmi".format(
                                    NfceRequestBuilder.NAMESPACE_NFE)).text
                                serie_nota = req.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}serie".format(
                                    NfceRequestBuilder.NAMESPACE_NFE)).text
                                serie_nota = serie_nota.zfill(3)
                                numero_nota = req.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}nNF".format(
                                    NfceRequestBuilder.NAMESPACE_NFE)).text
                                numero_nota = numero_nota.zfill(9)

                                self.nfce_situation_checker.logger = logger
                                fiscal_data = self.nfce_situation_checker.check_situation_nfe(nfe)
                                if not fiscal_data:
                                    message = "Problemas ao tentar consultar a situação da nota na sefaz. Vamos tentar novamente depois."
                                    raise Exception(message)

                                logger.info("Protocolo recebido. Order %s " % order_id)
                                fiscal_data_xml = eTree.XML(fiscal_data)

                                c_stat = fiscal_data_xml.find(
                                    ("{{{0}}}Body/{{{1}}}" + tag_consulta_nf_result + "/{{{2}}}retConsSitNFe/{{{2}}}cStat").format(
                                        NfceRequestBuilder.NAMESPACE_SOAP,
                                        namespace_consulta_situacao,
                                        NfceRequestBuilder.NAMESPACE_NFE))
                                if c_stat is None:
                                    raise Exception("Campo cStat Nao Encontrado")
                                else:
                                    c_stat = c_stat.text
                                if c_stat in ("100", "150", "613"):  # 100 = ok; 613 = Chave de Acesso difere da existente em BD;
                                    if c_stat in ("613", ):
                                        try:
                                            logger.info("Regerando XML. Order %s " % order_id)
                                            order_picture = eTree.XML(base64.b64decode(order_picture + "=" * ((4 - len(order_picture) % 4) % 4)))
                                            envelopado, data_emissao, serie_nota, numero_nota = self.fiscal_processor.nfce_request_builder.build_request(order_picture, False, None, None)
                                            serie_nota = str(serie_nota).zfill(3)
                                            numero_nota = str(numero_nota).zfill(9)
                                            logger.info("XML Regerado. Order %s " % order_id)
                                            index1 = envelopado.index("<NFe")
                                            index2 = envelopado.index("</NFe>")
                                            new_req = envelopado[index1:index2 + 6]

                                            request_str = request_str[:request_str.index("<NFe ")] + new_req + request_str[request_str.index("</NFe>") + 6:]
                                            req = eTree.XML(request_str)
                                        except Exception:
                                            logger.exception("Erro ao gerar XML da order %s" % order_id)
                                            fiscalrepository.set_nfce_sent(order_id, -1)
                                            continue

                                    if c_stat in ("613", ):
                                        nfe_key = \
                                        req.find("{{{0}}}NFe/{{{0}}}infNFe".format(NfceRequestBuilder.NAMESPACE_NFE)).attrib['Id']
                                        chave_xml = nfe_key[3:]

                                        self.nfce_situation_checker.logger = logger
                                        fiscal_data = self.nfce_situation_checker.check_situation_nfe(chave_xml)
                                        if not fiscal_data:
                                            message = "Problemas ao tentar consultar a situação da nota na sefaz. Vamos tentar novamente depois."
                                            raise Exception(message)

                                        logger.info("Protocolo recebido. Order %s " % order_id)
                                        fiscal_data_xml = eTree.XML(fiscal_data)

                                        c_stat = fiscal_data_xml.find(
                                            ("{{{0}}}Body/{{{1}}}"+ tag_consulta_nf_result + "/{{{2}}}retConsSitNFe/{{{2}}}cStat").format(
                                                NfceRequestBuilder.NAMESPACE_SOAP,
                                                namespace_consulta_situacao,
                                                NfceRequestBuilder.NAMESPACE_NFE))
                                        if c_stat is None:
                                            raise Exception("Campo cStat Nao Encontrado")
                                        else:
                                            c_stat = c_stat.text
                                        if c_stat not in ("100",):
                                            raise Exception("Campo cStat diferente de 100")

                                    inf_prot = fiscal_data_xml.find(
                                        ("{{{0}}}Body/{{{1}}}" + tag_consulta_nf_result + "/{{{2}}}retConsSitNFe/{{{2}}}protNFe/{{{2}}}infProt").format(
                                            NfceRequestBuilder.NAMESPACE_SOAP,
                                            namespace_consulta_situacao,
                                            NfceRequestBuilder.NAMESPACE_NFE))
                                    inf_prot.attrib.pop("Id", None)
                                    new_req = request_str[:request_str.index("<infProt")] + eTree.tostring(inf_prot) + request_str[request_str.index("</protNFe"):]
                                    xml_base64 = base64.b64encode(new_req)
                                    fiscalrepository.set_nfce_sent_with_xml_bkoffice(order_id, xml_base64, 1, 0)

                                    dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[5:7], data_emissao[8:10])
                                    dir_nota = os.path.join(self.fiscal_sent_dir, self.xml_enviados, dir_arquivo)
                                    self._create_directory_if_not_exists(dir_nota)
                                    nfce_file_name = "{0}_{1}_{2}_nfe_proc_pos{3}_{4}.xml".format(serie_nota, numero_nota, order_id.zfill(9), str(pos_id).zfill(2), nfe_key)

                                    for xml_file in glob.glob(dir_nota + '/*' + order_id.zfill(9) + '_nfe_proc_pos*'):
                                        os.remove(xml_file)

                                    with open(os.path.join(dir_nota, nfce_file_name), "w+") as nfe_proc_file:
                                        nfe_proc_file.write(new_req)
                                else:
                                    raise Exception("Status diferente de 100. %s" % c_stat)
                            except Exception:
                                logger.exception("Erro ao buscar protocolo da order %s" % order_id)
                                continue

                except Exception as _:
                    logger.exception("Erro ao Consultar Status dos XMLs")
                    return False, "Erro ao Consultar Status dos XMLs"
                finally:
                    if conn:
                        conn.close()

            return True, "Status Salvos com Sucesso"
        else:
            return False, "Loja nao opera com NFCE"

    def _handle_get_csts(self, _):
        return True, self.fiscal_parameter_controller.get_all_products_csts()

    def _handle_get_base_reductions(self, _):
        return True, self.fiscal_parameter_controller.get_all_products_base_reductions()

    def _handle_get_tip_params(self, _):
        return True, self.fiscal_parameter_controller.get_tip_params()

    def _handle_get_ibpt_tax(self, msg):
        order_xml = eTree.XML(msg.data)
        taxes_dict = self.ibpt_tax_processor.ibpt_tax_calculator(order_xml)
        return True, str(taxes_dict)

    def _handle_print_donation_coupon(self, msg):
        params = msg.data.split("\x00")
        report = self.donation_coupon_printer.print_coupon(*params)
        return True, report

    def _handle_get_contingency_status(self, _):
        return True, "Enabled" if self.nfce_contingencia.contingencia else "Disabled"

    def _handle_change_contingency_status(self, _):
        self.nfce_contingencia.force_contingency = not self.nfce_contingencia.force_contingency
        if self.nfce_contingencia.force_contingency:
            logger.info("Entrando em contingência forçada via menu suporte")
            self.nfce_contingencia.entra_contingencia("Contingencia forcada")
        else:
            self.nfce_contingencia.contingencia = False
            self.nfce_contingencia.process_now = True
            self.nfce_contingencia.start_contingencia()

        return True, "Contingency status successfully changed"

    def handle_event(self, subject, evt_type, data, msg):
        if subject == "ORDER_MODIFIED" and evt_type == "PAID":
            order = eTree.XML(data).find("Order")
            order_id = order.get("orderId")
            with FiscalDataRepository(self.mbcontext) as fiscal_repository:  # type: fiscalpersistence.FiscalDataRepository
                fiscal_repository.set_order_picture(order_id, base64.b64encode(eTree.tostring(order)))
        elif subject == "ORDER_MODIFIED" and evt_type == "VOID_ORDER":
            order = eTree.XML(data).find(".//Order")
            order_id = order.get("orderId")
            with FiscalDataRepository(self.mbcontext) as fiscal_repository:  # type: fiscalpersistence.FiscalDataRepository
                fiscal_repository.set_order_canceled(order_id)
        elif subject == "DisableNfceOrder":
            self._handle_disable_orders()
        elif subject == "ReSignXMLs":
            self._handle_re_sign_xml()
        elif subject == "OrderManagerFiscalXmlRequest":
            self._handle_fiscal_xml_request(data)
        elif subject == "CheckIfFiscalXmlIsCreatedInFolder":
            self._handle_check_if_fiscal_xml_is_created_in_folder()
        elif subject == "RetryFiscalOrdersWithException":
            self._handle_retry_fiscal_orders_with_exception()

    def _handle_search_request(self):
        if self.nf_type not in ("SAT", "MFE"):
            return False, "Loja não opera com SAT"

        sat_status_list = {}
        sat_service_finder = self.fiscal_processor.sat_service_finder  # type: SatServiceFinder

        pos_sat_list = sat_service_finder.pos_sat_list
        sat_list_to_search = pos_sat_list if pos_sat_list else range(0, sat_service_finder.max_sat_number + 1)

        for i in sat_list_to_search:
            sat_service_name = "SAT%s" % str(i).zfill(2)
            try:
                ret = self.mbcontext.MB_EasySendMessage(sat_service_name, TK_SAT_STATUS_REQUEST, format=FM_PARAM, timeout=100 * 100000)

                if ret.token == TK_SYS_ACK:
                    sat_status_list[sat_service_name] = ret.data
                else:
                    sat_status_list[sat_service_name] = "ERROR"
            except Exception as ex:
                logger.error("Erro buscando SAT - Exception: %s", str(ex))
                sat_status_list[sat_service_name] = "Desabilitado"

        return True, str(sat_status_list)

    def _handle_disable_orders(self):
        if not self.nf_type == "NFCE":
            return
        try:
            self.order_disabler.disable_all_orders(self.mbcontext)
        except Exception as _:
            logger.exception("Erro inutilizando notas")

    def _handle_sat_op_status_request(self, msg):
        pos_id = msg.data

        if self.nf_type not in ("SAT", "MFE"):
            return False, "Loja não opera com SAT"

        sat_service_finder = self.fiscal_processor.sat_service_finder  # type: SatServiceFinder
        pos_sat_list = sat_service_finder.pos_sat_list

        sat_info_list = []

        if not pos_sat_list or str(pos_id) in pos_sat_list:
            sat_service_name = "SAT%s" % pos_id.zfill(2)
            try:
                ret = self.mbcontext.MB_EasySendMessage(sat_service_name, TK_SAT_OPERATIONAL_STATUS_REQUEST, format=FM_PARAM, timeout=100 * 100000)
                sat_info_list.append(ret.data)
            except MBException as ex:
                logger.info("Error searching SAT on POS#{}: {}".format(pos_id, str(ex)))
                sat_info_list.append("ERROR - Desabilitado")
        else:
            sat_info_list.append("ERROR - Desabilitado")

        return True, str(sat_info_list)

    def _export_fiscal_files(self, msg):
        nok = False, 'Erro ao exportar arquivos fiscais'

        try:
            parse_ok, start_date, end_date = self._parse_date_from_fiscal_screen(msg.data)
            if not parse_ok:
                return nok

            try:
                folder_ok, folders = self._find_fiscal_folders(start_date, end_date, msg.data)
                if not folder_ok:
                    return nok
            except EmptySearchException as ex:
                return False, ex.message

            zip_ok, zip_path = self.zip_folders(folders)
            if not zip_ok:
                return nok

            return True, zip_path

        except Exception as ex:
            logger.error("Erro ao exportar arquivos fiscais %s", str(ex))

        return nok

    @staticmethod
    def _parse_date_from_fiscal_screen(dates):
        try:
            # parse das datas: yyyy/mm/dd;yyyy/mm/dd
            match = re.search(r'(\d+/\d+/\d+);(\d+/\d+/\d+)', dates)
            if match:
                dateSplit = dates.split(";")
                startDate = dateSplit[0]
                endDate = dateSplit[1]
                return True, startDate, endDate

        except Exception as ex:
            logger.error("Erro ao parsear datas (_parse_date_from_fiscal_screen) %s", str(ex))

        return False, '', ''

    def _find_fiscal_folders(self, start_date, end_date, msg):
        try:
            logger.info("start_date: " + start_date)
            logger.info("end_date: " + end_date)

            xml_type = self.xml_enviados if msg.split(";")[2] == '1' else self.xml_erros
            path = os.path.join(self.fiscal_sent_dir, xml_type)
            logger.info("path: " + path)

            export_folder = os.path.join(path, 'ExportTemp')
            logger.info("export_folder: " + export_folder)

            folders = filter(lambda f: os.path.isdir(f), glob.glob(path + '/*/*/*'))
            logger.info("lambda: " + str((lambda f: os.path.isdir(f), glob.glob(path + '/*/*/*'))))

            folders = [s.replace(path + "\\", '').replace(path + "/", '').replace('\\', '/') for s in folders]
            logger.info("folders after replace: " + str(folders))

            folders.sort()
            logger.info("folders after sort: " + str(folders))

            folders = folders[bisect.bisect_left(folders, start_date):bisect.bisect_right(folders, end_date)]
            logger.info("folders bisect_left: " + str(bisect.bisect_left(folders, start_date)))
            logger.info("folders bisect_right: " + str(bisect.bisect_right(folders, end_date)))
            logger.info("folders after bisect: " + str(folders))

            if folders:
                shutil.rmtree(export_folder, True)
                for item in folders:
                    shutil.copytree(os.path.join(path, item), os.path.join(export_folder, item))

                return True, export_folder
            else:
                raise EmptySearchException("Nenhum arquivo encontrado")

        except EmptySearchException as ex:
            raise ex

        except Exception as ex:
            logger.error("Erro ao buscar diretorios (_find_fiscal_folders) %s", str(ex))

        return False, ''

    def zip_folders(self, folder):
        try:
            zip_name = 'export'
            zip_complete_name = zip_name + '.zip'

            try:
                os.remove(os.path.join(self.export_xml_path, zip_complete_name))
            except OSError:
                pass

            self.zip(folder, zip_name)
            shutil.move(zip_complete_name, self.export_xml_path)
            shutil.rmtree(folder, True)
            return True, zip_complete_name
        except Exception as ex:
            logger.error("Erro ao zipar o diretorio temporario (zip_folders) %s", str(ex))
            shutil.rmtree(folder, True)

        return False, ''

    @staticmethod
    def zip(src, dst):
        zf = zipfile.ZipFile("%s.zip" % dst, "w", zipfile.ZIP_DEFLATED)
        abs_src = os.path.abspath(src)
        for dirname, subdirs, files in os.walk(src):
            for filename in files:
                absname = os.path.abspath(os.path.join(dirname, filename))
                arcname = absname[len(abs_src) + 1:]
                zf.write(absname, arcname)
        zf.close()

    def _handle_sat_process_payment_request(self, msg):
        if self.nf_type not in ("MFE",):
            return True, ''
        sat_service_name = self.fiscal_processor.sat_service_finder.find_and_lock_sat_service(msg.data.split("\0")[1])
        msg = self.mbcontext.MB_EasySendMessage(sat_service_name, TK_SAT_PROCESS_PAYMENT, format=FM_PARAM, data=msg.data)
        self.fiscal_processor.sat_service_finder.unlock_sat_service(sat_service_name)
        return True, msg.data

    def _handle_sat_payment_status_request(self, msg):
        if self.nf_type not in ("MFE",):
            return True, ''

        sat_service_name = self.fiscal_processor.sat_service_finder.find_and_lock_sat_service(msg.data.split("\0")[1])
        msg = self.mbcontext.MB_EasySendMessage(sat_service_name, TK_SAT_PAYMENT_STATUS, format=FM_PARAM, data=msg.data)
        self.fiscal_processor.sat_service_finder.unlock_sat_service(sat_service_name)
        return True, msg.data

    def _handle_cancel_order(self, msg):
        pos_id, order_id = msg.data.split("|")
        model = get_model(pos_id)
        order = eTree.XML(get_posot(model).orderPicture(pos_id, order_id))
        if order.tag == "Orders":
            order = order.find("Order")

        order_id = order.get('orderId')
        fiscalization_date = order.find(".//CustomOrderProperties/OrderProperty[@key='FISCALIZATION_DATE']")
        if fiscalization_date is None:
            return False, "Pedido ainda não fiscalizado"
        formatted_fiscalization_date = datetime.strptime(fiscalization_date.get("value"), "%Y-%m-%dT%H:%M:%S")
        since_datetime = datetime.now() - timedelta(minutes=30)
        if convert_from_utf_to_localtime(formatted_fiscalization_date).replace(tzinfo=None) < since_datetime:
            return False, "Pedido excedeu o tempo limite para cancelamento fiscal (30 minutos)"

        justificativa = "Teste de cancelamento"

        if self.nf_type == 'NFCE':
            with FiscalDataRepository(self.mbcontext) as fiscal_repository:
                order_data = fiscal_repository.get_xml_response(order_id)
                if order_data is None:
                    return False, "Só é possível cancelar vendas já autorizadas"
                status, data = self.nfce_canceler.cancel_order(order_data, justificativa, order_id)
                if status:
                    self.order_disabler.order_repository.mark_order_canceled(pos_id, order_id)
                    self.save_cancel_xml(pos_id, order_data)
                return status, data

        if self.nf_type in ['SAT', 'MFE']:
            with FiscalDataRepository(self.mbcontext) as fiscal_repository:
                order_data = fiscal_repository.get_numero_sat_and_xml_request(order_id)
                status, data = self.fiscal_processor.cancel_sale(order_data, order_id)
                if status:
                    self.save_cancel_xml(pos_id, order_data[1])

                if self.nf_type == 'MFE' and status and self.fiscal_processor.cancel_data:
                    try:
                        self.printers[""].print_cancel(pos_id, self.fiscal_processor.cancel_data)
                    except Exception as _:
                        logger.exception("Error print cancel fiscal data")

                return status, data

        return False, msg.data

    def save_cancel_xml(
            self,
            pos_id,     # type: str
            data        # type: str
    ):
        # type: (...) -> None

        with FiscalDataRepository(self.mbcontext) as fiscal_repository:
            fiscal_repository.set_fiscal_xml_custom_property(pos_id, data, key='CANCELED_FISCAL_XML')

    @staticmethod
    def _busca_parametros_sitef(order):
        # type: (eTree) -> (str, str, str)
        order_id = order.get("orderId")
        created_at = order.get("createdAt").replace("-", "").replace(":", "")
        data_fiscal = created_at[:8]
        hora_fiscal = created_at[9:15]

        return order_id, data_fiscal, hora_fiscal

    def _reprint_sat_or_nfce(self, xml_request, xml_response, params, danfe_type):
        if not xml_request:
            return False, "XML do pedido não encontrado"
        s = base64.b64decode(xml_request)
        contingencia = s[s.find("<tpEmis>")+8:s.find("</tpEmis>")]
        if contingencia == "9":
            return False, "Nota ainda não autorizada pela SEFAZ"
        if self.nf_type in ("SAT", "MFE"):
            ret_xml_request = self._get_xml_sat(xml_request)
            if None in [ret_xml_request['fiscal_data'], ret_xml_request['xml_request']]:
                return False, "Erro ao obter XML da venda!"
            if danfe_type in self.printers:
                selected_printer = self.printers[danfe_type]
            else:
                selected_printer = self.printers[""]
            return True, selected_printer.print_fiscal(params.get('posid'),
                                                       params.get('order'),
                                                       params.get('tenders'),
                                                       ret_xml_request.get("xml_request", ""),
                                                       eTree.XML(ret_xml_request.get("fiscal_data", "")),
                                                       params.get("total_taxes"),
                                                       params.get('customer_name'))
        else:
            try:
                ret_xml_request = self._get_xml_nfce(xml_request, xml_response)
            except StopIteration as ex:
                return False, ex.message
            if None in [ret_xml_request['request_xml'], ret_xml_request['response_xml']]:
                return False, "Erro ao obter XML da venda!"
            if danfe_type in self.printers:
                selected_printer = self.printers[danfe_type]
            else:
                selected_printer = self.printers[""]
            return True, selected_printer.print_fiscal(params.get('posid'),
                                                       params.get('order'),
                                                       params.get('tenders'),
                                                       ret_xml_request.get('request_xml'),
                                                       ret_xml_request.get('response_xml'),
                                                       params.get("total_taxes"),
                                                       params.get('customer_name'),
                                                       params.get('url_sefaz'))

    @staticmethod
    def _get_xml_sat(xml_request):
        try:
            xml = eTree.XML(base64.b64decode(xml_request))
            ide = xml.find('infCFe').find('ide')
            sat_key = xml.find('infCFe').attrib.get('Id')

            cpf_cnpj = xml.find('infCFe').find('dest') or ""
            if cpf_cnpj:
                if cpf_cnpj.find("CPF") is not None:
                    cpf_cnpj = cpf_cnpj.find("CPF").text
                elif cpf_cnpj.find("CNPJ") is not None:
                    cpf_cnpj = cpf_cnpj.find("CNPJ").text
                else:
                    cpf_cnpj = None

            date_time = "{0}{1}".format(ide.find('dEmi').text, ide.find('hEmi').text)
            date_time_reprint = datetime.strptime(date_time, "%Y%m%d%H%M%S").strftime("%d/%m/%y %H:%M:%S")

            total = xml.find('infCFe').find('total')

            customer_doc = cpf_cnpj if cpf_cnpj is not None else 'Nao informado'

            qr_code_values = [sat_key[3:],
                              date_time,
                              total.find('vCFe').text,
                              customer_doc,
                              ide.find('assinaturaQRCODE').text]

            qr_code = "|".join(qr_code_values)

            list_tags_xml = ['vICMS', 'vPIS', 'vCOFINS', 'vPISST', 'vCOFINSST']
            total_taxes = 0
            icms_total = total.find('ICMSTot')
            for x in list_tags_xml:
                try:
                    total_taxes += D(icms_total.find(x).text)
                except (TypeError, AttributeError):
                    pass
            total_taxes = round_half_away_from_zero(total_taxes, 2)

            tag_xml_request = eTree.XML("<XmlRequest>{0}</XmlRequest>".format(xml_request))
            tag_fiscal_data = "<FiscalData>"
            tag_fiscal_data += "<datetime>{0}</datetime>".format(date_time_reprint)
            tag_fiscal_data += "<satno>{0}</satno>".format(ide.find('nserieSAT').text)
            tag_fiscal_data += "<satkey>{0}</satkey>".format(sat_key[3:])
            tag_fiscal_data += "<CustomerCPF>{0}</CustomerCPF>".format(customer_doc)
            tag_fiscal_data += "<QRCode>{0}</QRCode>".format(qr_code)
            tag_fiscal_data += "</FiscalData>"

            return {"xml_request": tag_xml_request, "fiscal_data": tag_fiscal_data, "total_taxes": total_taxes}
        except Exception as ex:
            logger.error("Erro ao construir o xml - Exception %s", str(ex))
            return {"xml_request": None, "fiscal_data": None, "total_taxes": 0}

    def _get_xml_nfce(self, xml_request, xml_response):
        try:
            xml = remove_xml_namespace(base64.b64decode(xml_request))
            xml_resp = remove_xml_namespace(base64.b64decode(xml_response))

            response_tag = "retConsReciNFe"
            if self.synchronous_mode:
                response_tag = "retEnviNFe"
            xml_ws_version = int(float(xml_resp.find(".//{}".format(response_tag)).get('versao')))

            if xml_ws_version != self.versao_ws:
                raise StopIteration("Versão do XML diferente da versão fiscal atual. Versão XML: {}; Versão Fiscal: {}"
                                    .format(xml_ws_version, self.versao_ws))

            nprot = xml_resp.find(".//infProt/nProt").text
            data_resp = xml_resp.find(".//infProt/dhRecbto").text

            if xml.find('.//NFe'):
                body = eTree.tostring(xml.find('.//NFe'))
            else:
                body = eTree.tostring(xml)

            version = xml.find(".//infNFe").attrib['versao']
            total = xml.find('.//ICMSTot')

            list_tags_xml = ['vICMS', 'vPIS', 'vCOFINS', 'vPISST', 'vCOFINSST']
            total_taxes = 0

            for tag in list_tags_xml:
                try:
                    total_taxes += D(total.find('.//{}'.format(tag)).text)
                except (TypeError, AttributeError):
                    pass

            total_taxes = round_half_away_from_zero(total_taxes, 2)

            tag_request = '<ns0:Envelope xmlns = "http://www.portalfiscal.inf.br/nfe" ' \
                          'xmlns:ns0 = "http://www.w3.org/2003/05/soap-envelope" ' \
                          'xmlns:ns1 = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeAutorizacao" ' \
                          'xmlns:ns3 = "http://www.w3.org/2000/09/xmldsig#">'
            tag_request += '<ns0:Body><ns1:nfeDadosMsg>'
            tag_request += '<enviNFe versao = "{0}">'.format(version)
            tag_request += '{0}'.format(body)
            tag_request += '</enviNFe></ns1:nfeDadosMsg></ns0:Body></ns0:Envelope>'

            tag_reponse = '<ns0:Envelope xmlns="http://www.portalfiscal.inf.br/nfe" ' \
                          'xmlns:ns0="http://www.w3.org/2003/05/soap-envelope" ' \
                          'xmlns:ns1="http://www.portalfiscal.inf.br/nfe/wsdl/NfeRetAutorizacao">' \
                          '<ns0:Body><ns1:nfeRetAutorizacaoLoteResult>'
            tag_reponse += '<retConsReciNFe versao="{0}">'.format(version)
            tag_reponse += '<protNFe versao="{0}"><infProt>'.format(version)
            tag_reponse += '<dhRecbto>{0}</dhRecbto>'.format(str(data_resp))
            tag_reponse += '<nProt>{0}</nProt>'.format(str(nprot))
            tag_reponse += '</infProt></protNFe></retConsReciNFe></ns1:nfeRetAutorizacaoLoteResult></ns0:Body></ns0:Envelope>'

            return {"request_xml": eTree.XML(tag_request), "response_xml": eTree.XML(tag_reponse), "total_taxes": total_taxes}
        except StopIteration:
            raise
        except Exception as ex:
            logger.error("Erro ao construir o xml - Exception %s", str(ex))
            return {"request_xml": None, "response_xml": None, "total_taxes": 0}

    def _handle_re_sign_xml(self):
        with FiscalDataRepository(self.mbcontext) as fiscalrepository:
            logger.info("Starting XML Resign Event...")
            conn = None
            try:
                conn = DBDriver().open(self.mbcontext, service_name="FiscalPersistence")
                sql = """SELECT PosId, OrderId, OrderPicture, XMLRequest
                          FROM fiscal.FiscalData
                          WHERE SentToNfce IN (555)
                          ORDER BY OrderId ASC LIMIT 500"""

                cursor = conn.select(sql)
                orders_with_error = 0
                orders_with_success = 0
                for row in cursor:
                    try:
                        pos_id, order_id, order_picture, request = map(row.get_entry, ("PosId", "OrderId", "OrderPicture", "XMLRequest"))
                        logger.info("Resigning Order %s" % order_id)

                        if self.nf_type != "PAF":
                            try:
                                is_in_contingence = True
                                xml_decoded = base64.b64decode(request + "=" * ((4 - len(request) % 4) % 4))
                                req = remove_xml_namespace(xml_decoded)

                                dh_cont = req.find(".//infNFe/ide/dhCont")
                                if dh_cont is None:
                                    dh_cont = req.find(".//NFe/infNFe/ide/dhCont")

                                x_just = req.find(".//infNFe/ide/xJust")
                                if x_just is None:
                                    x_just = req.find(".//NFe/infNFe/ide/xJust")

                                if dh_cont is not None:
                                    dh_cont = dh_cont.text
                                    dh_cont = dh_cont[:-6]
                                    dh_cont = datetime.strptime(dh_cont, "%Y-%m-%dT%H:%M:%S")
                                if x_just is not None:
                                    x_just = x_just.text
                                if dh_cont is None and x_just is None:
                                    dh_cont = datetime.now()
                                    x_just = "Problemas de conexao com a SEFAZ"
                            except Exception as _:
                                dh_cont = datetime.now()
                                x_just = "Problemas de conexao com a SEFAZ"

                                logger.info("Order: {} - dh_cont: {} / xjust: {}".format(order_id, dh_cont, x_just))
                        else:
                            is_in_contingence = False
                            dh_cont = ""
                            x_just = ""
                            logger.info("Order: {} - Not in contingence".format(order_id))

                        order_taker = OrderTaker()
                        order = order_taker.get_order_picture(pos_id, order_id)
                        fiscalrepository.set_order_picture(order_id, base64.b64encode(order))
                        order = eTree.XML(order)

                        order.attrib["posId"] = pos_id

                        fiscal_data = self.fiscal_processor.nfce_request_builder.build_request(order, is_in_contingence, dh_cont, x_just)
                        index1 = fiscal_data[0].index("<NFe")
                        index2 = fiscal_data[0].index("</NFe>")
                        new_req = fiscal_data[0][index1:index2 + 6]
                        status = 0
                        xml_base64 = base64.b64encode(new_req)

                        fiscalrepository.set_nfce_sent_with_xml(order_id, xml_base64, status)
                        logger.info("XML Resign Order %s OK" % order_id)
                        orders_with_success += 1

                    except Exception as _:
                        orders_with_error += 1
                        logger.exception("Error to build XML of the order %s" % order_id)
                        continue
                if orders_with_error:
                    logger.error("Orders with success: {0}. Orders with error: {1}".format(orders_with_success, orders_with_error))
                    return False, "Orders with success: {0}. Orders with error: {1}".format(orders_with_success, orders_with_error)

            except Exception as _:
                logger.exception("Error to resign XMLs")
                return False, "Error to resign XMLs"
            finally:
                logger.info("Finish XML Resign Event...")
                if conn:
                    conn.close()

        return True, "Success to resigns XMLs. Number of orders resigns: {}".format(orders_with_success)

    def _handle_fiscal_xml_request(self, data, return_data=False):
        conn = None
        base64_xml_response = None
        order_id = None
        try:
            conn = DBDriver().open(self.mbcontext, service_name="FiscalPersistence")
            sql = """SELECT OrderId, XMLRequest
                      FROM fiscal.FiscalData
                      WHERE OrderId=%s""" % data
            cursor = conn.select(sql)
            for row in cursor:
                order_id = row.get_entry(0)
                base64_xml_response = row.get_entry(1)
        except Exception:
            logger.exception("Erro ao buscar XMLRequest")
        finally:
            if conn:
                conn.close()
        if return_data:
            if order_id is None:
                return False, "Invalid Order ID {}".format(data)
            if base64_xml_response is None:
                return False, "No XML for order ID {}".format(order_id)
            return True, base64_xml_response
        else:
            if order_id is None:
                self.mbcontext.MB_EasyEvtSend("PosFiscalXmlError", "", ";".join((order_id, "OrderId inválida ou inexistente no banco - order %s" % data)))
            elif base64_xml_response is None:
                self.mbcontext.MB_EasyEvtSend("PosFiscalXmlError", "", ";".join((order_id, "XML inexistente para order %s" % order_id)))
            else:
                xml_response = base64.b64decode(base64_xml_response)
                compress_obj = zlib.compressobj(zlib.Z_BEST_COMPRESSION, zlib.DEFLATED, 31)
                zipped_xml = compress_obj.compress(xml_response)
                zipped_xml += compress_obj.flush()
                base64_zipped_xml = base64.b64encode(zipped_xml)
                self.mbcontext.MB_EasyEvtSend("PosFiscalXmlResponse", "", ";".join((order_id, base64_zipped_xml)))

    def _handle_check_if_fiscal_xml_is_created_in_folder(self):
        # type: () -> None
        try:
            logger.info("Starting -> _handle_check_if_fiscal_xml_is_created_in_folder")
            date_yesterday, date_yesterday_start, date_yesterday_end = self._get_yesterday_dates()
            fiscal_data_path = self._get_fiscal_xml_sent_path(date_yesterday)

            database_fiscal_data_number_list = self._get_fiscal_data_number_from_database_by_date(date_yesterday_start, date_yesterday_end)

            self._generate_fiscal_xml_not_created(database_fiscal_data_number_list, fiscal_data_path)
            logger.info("Finished -> _handle_check_if_fiscal_xml_is_created_in_folder")
        except Exception as _:
            logger.exception("Error creating XML files")

    def _handle_retry_fiscal_orders_with_exception(self):
        # type: () -> None
        try:
            logger.info("Starting -> _handle_retry_fiscal_orders_with_exception")
            with FiscalDataRepository(self.mbcontext) as fiscal_repository:  # type: fiscalpersistence.FiscalDataRepository
                fiscal_repository.retry_fiscal_orders_with_exception(self.period_to_retry_orders_with_exception)

            logger.info("Finished -> _handle_retry_fiscal_orders_with_exception")
        except Exception as _:
            logger.exception("Error changing status of fiscal orders with exception")

    def _get_fiscal_xml_sent_path(self, date_yesterday):
        fiscal_data_file = str(date_yesterday.year) + '/' + str(date_yesterday.month).zfill(2) + '/' + str(date_yesterday.day).zfill(2)
        fiscal_data_path = os.path.join(self.fiscal_sent_dir, self.xml_enviados, fiscal_data_file)
        return fiscal_data_path

    def _generate_fiscal_xml_not_created(self, database_fiscal_data_number_list, fiscal_data_path):
        self._create_directory_if_not_exists(fiscal_data_path)

        self._verify_if_xml_in_folder_is_in_database_list(database_fiscal_data_number_list, fiscal_data_path)
        self._create_xml_if_it_is_not_in_folder(database_fiscal_data_number_list, fiscal_data_path)

    def _create_xml_if_it_is_not_in_folder(self, database_fiscal_data_number_list, fiscal_data_path):
        conn = None
        try:
            conn = DBDriver().open(self.mbcontext, service_name="FiscalPersistence")
            sql = """SELECT PosId, OrderId, XMLRequest, NumeroNota, NumeroSat FROM FiscalData
                     WHERE NumeroNota IN ({}) """.format("'" + "','".join(database_fiscal_data_number_list) + "'")

            cursor = conn.select(sql)
            for row in cursor:
                numero_nota, order_id, pos_id, xml_request, numero_sat = self._parse_database_info(row)

                inf_id, numero_serie = self._parse_xml_info(xml_request)

                fiscal_xml_path_to_save = self.create_fiscal_xml_path_to_save(inf_id, numero_serie,
                                                                              fiscal_data_path, numero_nota, order_id,
                                                                              pos_id, numero_sat)
                self._save_xml_file_in_folder(fiscal_xml_path_to_save, xml_request)
        except Exception as ex:
            logger.exception("Exception creating XML with error: {}".format(ex))
        finally:
            conn.close()

    def create_fiscal_xml_path_to_save(self, inf_id, numero_serie, fiscal_data_path, numero_nota, order_id, pos_id, numero_sat):
        if self.nf_type in ["NFCE", "PAF"]:
            fiscal_xml_path_to_save = FiscalWrapperEventHandler._generate_nfce_xml_file_name(inf_id,
                                                                                             numero_nota,
                                                                                             order_id,
                                                                                             pos_id,
                                                                                             numero_serie,
                                                                                             fiscal_data_path)
        else:

            fiscal_xml_path_to_save = FiscalWrapperEventHandler._generate_sat_xml_file_name(fiscal_data_path,
                                                                                            numero_nota,
                                                                                            order_id,
                                                                                            pos_id,
                                                                                            numero_sat,
                                                                                            inf_id)
        return fiscal_xml_path_to_save

    @staticmethod
    def _generate_nfce_xml_file_name(infnfe_id, numero_nota, order_id, pos_id, numero_serie, fiscal_data_path):
        fiscal_xml_path_to_save = os.path.join(fiscal_data_path, "{0}_{1}_{2}_nfe_proc_pos{3}_{4}.xml".format(
            numero_serie.zfill(3),
            numero_nota.zfill(9),
            order_id.zfill(9),
            pos_id.zfill(2),
            infnfe_id))
        return fiscal_xml_path_to_save

    @staticmethod
    def _generate_sat_xml_file_name(fiscal_data_path, numero_nota, order_id, pos_id, numero_sat, infcfe_id):
        fiscal_xml_path_to_save = os.path.join(fiscal_data_path, "S{0}_{1}_{2}_response_pos{3}_{4}.xml".format(
            numero_sat[-2:],
            str(numero_nota).zfill(9),
            str(order_id).zfill(9),
            str(pos_id).zfill(2),
            infcfe_id))
        return fiscal_xml_path_to_save

    @staticmethod
    def _save_xml_file_in_folder(fiscal_xml_path_to_save, fiscal_xml_to_save):
        with open(fiscal_xml_path_to_save, "w+") as xml_to_save:
            xml_to_save.write(fiscal_xml_to_save)

    def _parse_xml_info(self, fiscal_xml_to_save):
        fiscal_xml_to_save = remove_xml_namespace(fiscal_xml_to_save)

        key_fiscal_inf = fiscal_xml_to_save.find(".//infNFe") if self.nf_type in ("NFCE", "PAF") else fiscal_xml_to_save.find(".//infCFe")
        data_fiscal_inf = key_fiscal_inf.attrib['Id']

        key_data_fiscal_serie = fiscal_xml_to_save.find(".//serie") if self.nf_type in ("NFCE", "PAF") else fiscal_xml_to_save.find(".//nserieSAT")
        data_fiscal_serie = key_data_fiscal_serie.text.zfill(3)

        return data_fiscal_inf, data_fiscal_serie

    @staticmethod
    def _parse_database_info(row):
        numero_nota = str(row.get_entry("NumeroNota")).zfill(9)
        order_id = str(row.get_entry("OrderId")).zfill(9)
        pos_id = str(row.get_entry("PosId")).zfill(2)
        xml_request = base64.b64decode(row.get_entry("XMLRequest"))
        numero_sat = str(row.get_entry("NumeroSat"))
        return numero_nota, order_id, pos_id, xml_request, numero_sat

    def _verify_if_xml_in_folder_is_in_database_list(self, database_fiscal_data_number_list, fiscal_data_path):
        for xml_file in glob.glob(os.path.join(fiscal_data_path, '*.xml')):
            try:
                xml_fiscal_number = self._get_xml_fiscal_number_from_xml_file(xml_file)

                if xml_fiscal_number in database_fiscal_data_number_list:
                    database_fiscal_data_number_list.remove(xml_fiscal_number)
                elif not 'procInut' in xml_file:
                    self._move_xml_to_inconsistent_folder(fiscal_data_path, xml_file)

            except Exception as _:
                logger.exception("Exception on File: " + xml_file + "\nMoved to folder: " + self.xml_inconsistentes)
                self._move_xml_to_inconsistent_folder(fiscal_data_path, xml_file)

    def _get_xml_fiscal_number_from_xml_file(self, xml_file):
        xml = remove_xml_namespace(eTree.tostring(eTree.parse(xml_file).getroot()))
        key = xml.find(".//nNF") if self.nf_type == "NFCE" else xml.find(".//nCFe")
        return key.text if key is not None else None

    def _move_xml_to_inconsistent_folder(self, fiscal_data_path, xml_file):
        inconsistent_fiscal_data_path = fiscal_data_path.replace(self.xml_enviados, self.xml_inconsistentes)
        self._create_directory_if_not_exists(inconsistent_fiscal_data_path)
        xml_file_new_path = xml_file.replace(self.xml_enviados, self.xml_inconsistentes)
        shutil.move(xml_file, xml_file_new_path)
        logger.info("File: " + xml_file + "\nMoved to folder: " + self.xml_inconsistentes)

    @staticmethod
    def _create_directory_if_not_exists(data_path):
        try:
            if not os.path.exists(data_path):
                os.makedirs(data_path)
        except OSError:
            logger.error("Cannot create path: {}".format(data_path))

    @staticmethod
    def _get_yesterday_dates():
        date_yesterday = datetime.now() - timedelta(days=1)
        date_yesterday_start = datetime(date_yesterday.year, date_yesterday.month, date_yesterday.day, 0, 0, 0)
        date_yesterday_end = datetime(date_yesterday.year, date_yesterday.month, date_yesterday.day, 23, 59, 59)
        return date_yesterday, date_yesterday_start, date_yesterday_end

    def _get_fiscal_data_number_from_database_by_date(self, start_date, end_date):
        # type: (datetime, datetime) -> [unicode]
        fiscal_data_number_list = []
        conn = None

        try:
            conn = DBDriver().open(self.mbcontext, service_name="FiscalPersistence")
            sql = """SELECT NumeroNota, datetime(datanota,'unixepoch','localtime') AS data 
                     FROM fiscal.FiscalData
                     WHERE data > '{}' AND data <= '{}'""".format(start_date, end_date)

            cursor = conn.select(sql)
            for row in cursor:
                fiscal_data_number_list.append(row.get_entry("NumeroNota"))
        except Exception as ex:
            logger.exception("Get fiscal data number list error: " + ex.message)
        finally:
            conn.close()

        return fiscal_data_number_list



from production.model.request import ChangePathRequest

from ._ChangePathProcessor import ChangePathProcessor


class EnablePathProcessor(ChangePathProcessor):
    def parse_data(self, data):
        return ChangePathRequest(data, True)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function


from cryptography import utils
from cryptography.exceptions import (
    InvalidSignature, UnsupportedAlgorithm, _Reasons
)
from cryptography.hazmat.primitives import constant_time
from cryptography.hazmat.primitives.ciphers.modes import CBC


class _CMACContext(object):
    def __init__(self, backend, algorithm, ctx=None):
        if not backend.cmac_algorithm_supported(algorithm):
            raise UnsupportedAlgorithm("This backend does not support CMAC.",
                                       _Reasons.UNSUPPORTED_CIPHER)

        self._backend = backend
        self._key = algorithm.key
        self._algorithm = algorithm
        self._output_length = algorithm.block_size // 8

        if ctx is None:
            registry = self._backend._cipher_registry
            adapter = registry[type(algorithm), CBC]

            evp_cipher = adapter(self._backend, algorithm, CBC)

            ctx = self._backend._lib.CMAC_CTX_new()

            self._backend.openssl_assert(ctx != self._backend._ffi.NULL)
            ctx = self._backend._ffi.gc(ctx, self._backend._lib.CMAC_CTX_free)

            key_ptr = self._backend._ffi.from_buffer(self._key)
            res = self._backend._lib.CMAC_Init(
                ctx, key_ptr, len(self._key),
                evp_cipher, self._backend._ffi.NULL
            )
            self._backend.openssl_assert(res == 1)

        self._ctx = ctx

    algorithm = utils.read_only_property("_algorithm")

    def update(self, data):
        res = self._backend._lib.CMAC_Update(self._ctx, data, len(data))
        self._backend.openssl_assert(res == 1)

    def finalize(self):
        buf = self._backend._ffi.new("unsigned char[]", self._output_length)
        length = self._backend._ffi.new("size_t *", self._output_length)
        res = self._backend._lib.CMAC_Final(
            self._ctx, buf, length
        )
        self._backend.openssl_assert(res == 1)

        self._ctx = None

        return self._backend._ffi.buffer(buf)[:]

    def copy(self):
        copied_ctx = self._backend._lib.CMAC_CTX_new()
        copied_ctx = self._backend._ffi.gc(
            copied_ctx, self._backend._lib.CMAC_CTX_free
        )
        res = self._backend._lib.CMAC_CTX_copy(
            copied_ctx, self._ctx
        )
        self._backend.openssl_assert(res == 1)
        return _CMACContext(
            self._backend, self._algorithm, ctx=copied_ctx
        )

    def verify(self, signature):
        digest = self.finalize()
        if not constant_time.bytes_eq(digest, signature):
            raise InvalidSignature("Signature did not match digest.")



class DeliveryBodyDto(object):
    def __init__(self, order):
        # type: (object) -> ()
        self.items = order.items
        self.total_price = order.total_price
        self.sub_total_price = order.sub_total_price
        self.delivery_fee = order.delivery_fee
        self.observation = order.observation
        self.observations = order.observations
        self.tenders = order.tenders
        self.customer_name = order.customer_name
        self.customer_doc = order.customer_doc
        self.customer_phone = order.customer_phone
        self.address = order.address
        self.address_neighborhood = order.address_neighborhood
        self.city = order.city
        self.address_complement = order.address_complement
        self.postal_code = order.postal_code
        self.vouchers = order.vouchers



"""
Serve Shared Static Files
=========================

.. autoclass:: SharedDataMiddleware
    :members: is_allowed

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""
import mimetypes
import os
import pkgutil
import posixpath
from datetime import datetime
from io import BytesIO
from time import mktime
from time import time
from zlib import adler32

from .._compat import PY2
from .._compat import string_types
from ..filesystem import get_filesystem_encoding
from ..http import http_date
from ..http import is_resource_modified
from ..security import safe_join
from ..utils import get_content_type
from ..wsgi import get_path_info
from ..wsgi import wrap_file


class SharedDataMiddleware(object):

    """A WSGI middleware that provides static content for development
    environments or simple server setups. Usage is quite simple::

        import os
        from werkzeug.middleware.shared_data import SharedDataMiddleware

        app = SharedDataMiddleware(app, {
            '/static': os.path.join(os.path.dirname(__file__), 'static')
        })

    The contents of the folder ``./shared`` will now be available on
    ``http://example.com/shared/``.  This is pretty useful during development
    because a standalone media server is not required.  One can also mount
    files on the root folder and still continue to use the application because
    the shared data middleware forwards all unhandled requests to the
    application, even if the requests are below one of the shared folders.

    If `pkg_resources` is available you can also tell the middleware to serve
    files from package data::

        app = SharedDataMiddleware(app, {
            '/static': ('myapplication', 'static')
        })

    This will then serve the ``static`` folder in the `myapplication`
    Python package.

    The optional `disallow` parameter can be a list of :func:`~fnmatch.fnmatch`
    rules for files that are not accessible from the web.  If `cache` is set to
    `False` no caching headers are sent.

    Currently the middleware does not support non ASCII filenames.  If the
    encoding on the file system happens to be the encoding of the URI it may
    work but this could also be by accident.  We strongly suggest using ASCII
    only file names for static files.

    The middleware will guess the mimetype using the Python `mimetype`
    module.  If it's unable to figure out the charset it will fall back
    to `fallback_mimetype`.

    :param app: the application to wrap.  If you don't want to wrap an
                application you can pass it :exc:`NotFound`.
    :param exports: a list or dict of exported files and folders.
    :param disallow: a list of :func:`~fnmatch.fnmatch` rules.
    :param cache: enable or disable caching headers.
    :param cache_timeout: the cache timeout in seconds for the headers.
    :param fallback_mimetype: The fallback mimetype for unknown files.

    .. versionchanged:: 1.0
        The default ``fallback_mimetype`` is
        ``application/octet-stream``. If a filename looks like a text
        mimetype, the ``utf-8`` charset is added to it.

    .. versionadded:: 0.6
        Added ``fallback_mimetype``.

    .. versionchanged:: 0.5
        Added ``cache_timeout``.
    """

    def __init__(
        self,
        app,
        exports,
        disallow=None,
        cache=True,
        cache_timeout=60 * 60 * 12,
        fallback_mimetype="application/octet-stream",
    ):
        self.app = app
        self.exports = []
        self.cache = cache
        self.cache_timeout = cache_timeout

        if hasattr(exports, "items"):
            exports = exports.items()

        for key, value in exports:
            if isinstance(value, tuple):
                loader = self.get_package_loader(*value)
            elif isinstance(value, string_types):
                if os.path.isfile(value):
                    loader = self.get_file_loader(value)
                else:
                    loader = self.get_directory_loader(value)
            else:
                raise TypeError("unknown def %r" % value)

            self.exports.append((key, loader))

        if disallow is not None:
            from fnmatch import fnmatch

            self.is_allowed = lambda x: not fnmatch(x, disallow)

        self.fallback_mimetype = fallback_mimetype

    def is_allowed(self, filename):
        """Subclasses can override this method to disallow the access to
        certain files.  However by providing `disallow` in the constructor
        this method is overwritten.
        """
        return True

    def _opener(self, filename):
        return lambda: (
            open(filename, "rb"),
            datetime.utcfromtimestamp(os.path.getmtime(filename)),
            int(os.path.getsize(filename)),
        )

    def get_file_loader(self, filename):
        return lambda x: (os.path.basename(filename), self._opener(filename))

    def get_package_loader(self, package, package_path):
        loadtime = datetime.utcnow()
        provider = pkgutil.get_loader(package)

        if hasattr(provider, "get_resource_reader"):
            # Python 3
            reader = provider.get_resource_reader(package)

            def loader(path):
                if path is None:
                    return None, None

                path = safe_join(package_path, path)
                basename = posixpath.basename(path)

                try:
                    resource = reader.open_resource(path)
                except IOError:
                    return None, None

                if isinstance(resource, BytesIO):
                    return (
                        basename,
                        lambda: (resource, loadtime, len(resource.getvalue())),
                    )

                return (
                    basename,
                    lambda: (
                        resource,
                        datetime.utcfromtimestamp(os.path.getmtime(resource.name)),
                        os.path.getsize(resource.name),
                    ),
                )

        else:
            # Python 2
            package_filename = provider.get_filename(package)
            is_filesystem = os.path.exists(package_filename)
            root = os.path.join(os.path.dirname(package_filename), package_path)

            def loader(path):
                if path is None:
                    return None, None

                path = safe_join(root, path)
                basename = posixpath.basename(path)

                if is_filesystem:
                    if not os.path.isfile(path):
                        return None, None

                    return basename, self._opener(path)

                try:
                    data = provider.get_data(path)
                except IOError:
                    return None, None

                return basename, lambda: (BytesIO(data), loadtime, len(data))

        return loader

    def get_directory_loader(self, directory):
        def loader(path):
            if path is not None:
                path = safe_join(directory, path)
            else:
                path = directory

            if os.path.isfile(path):
                return os.path.basename(path), self._opener(path)

            return None, None

        return loader

    def generate_etag(self, mtime, file_size, real_filename):
        if not isinstance(real_filename, bytes):
            real_filename = real_filename.encode(get_filesystem_encoding())

        return "wzsdm-%d-%s-%s" % (
            mktime(mtime.timetuple()),
            file_size,
            adler32(real_filename) & 0xFFFFFFFF,
        )

    def __call__(self, environ, start_response):
        path = get_path_info(environ)

        if PY2:
            path = path.encode(get_filesystem_encoding())

        file_loader = None

        for search_path, loader in self.exports:
            if search_path == path:
                real_filename, file_loader = loader(None)

                if file_loader is not None:
                    break

            if not search_path.endswith("/"):
                search_path += "/"

            if path.startswith(search_path):
                real_filename, file_loader = loader(path[len(search_path) :])

                if file_loader is not None:
                    break

        if file_loader is None or not self.is_allowed(real_filename):
            return self.app(environ, start_response)

        guessed_type = mimetypes.guess_type(real_filename)
        mime_type = get_content_type(guessed_type[0] or self.fallback_mimetype, "utf-8")
        f, mtime, file_size = file_loader()

        headers = [("Date", http_date())]

        if self.cache:
            timeout = self.cache_timeout
            etag = self.generate_etag(mtime, file_size, real_filename)
            headers += [
                ("Etag", '"%s"' % etag),
                ("Cache-Control", "max-age=%d, public" % timeout),
            ]

            if not is_resource_modified(environ, etag, last_modified=mtime):
                f.close()
                start_response("304 Not Modified", headers)
                return []

            headers.append(("Expires", http_date(time() + timeout)))
        else:
            headers.append(("Cache-Control", "public"))

        headers.extend(
            (
                ("Content-Type", mime_type),
                ("Content-Length", str(file_size)),
                ("Last-Modified", http_date(mtime)),
            )
        )
        start_response("200 OK", headers)
        return wrap_file(environ, f)
